<?php defined('ABSPATH') or die("you do not have access to this page!");

class rsssl_headers {

	private static $_this;
	public $cors_headers;

	function __construct()
	{
		if (isset(self::$_this))
			wp_die(sprintf(__('%s is a singleton class and you cannot create a second instance.', 'really-simple-ssl'), get_class($this)));

		self::$_this = $this;
        if ( rsssl_get_networkwide_option('rsssl_security_headers_method') === 'advancedheaders') {
			add_action( "admin_init", array( $this, "insert_security_headers" ), 90 );
	        add_filter( 'rocket_htaccess_mod_rewrite', '__return_false' );
        }

		if ( $this->is_settings_page() ) {
			add_action( "update_option_rsssl_security_headers_method", array( $this, "update_security_headers_method_functions" ), 10, 3 );

			delete_transient('rsssl_admin_notices');
			//start new check right away
			//without parameter clears data
//@todo Header detection disabled, enable in future version
			//add_action('plugins_loaded', array($this, 'get_detected_security_headers') );
		}
	}

	static function this()
	{
		return self::$_this;
	}

	/**
	 * Remove advanced-headers.php if option is set to .htaccess
	 * @since 5.5
	 */

	public function update_security_headers_method_functions( $old_value, $value, $option) {
		if ( $old_value==$value ) {
			return;
		}

		if ( RSSSL_PRO()->rsssl_premium_options->get_networkwide_option('rsssl_security_headers_method') !== 'advancedheaders' ) {
			rsssl_pro_remove_advanced_headers();
		}

		//of headers method change, ensure header check renews by redirecting
		wp_redirect(add_query_arg(array('page' => 'rlrsssl_really_simple_ssl', 'tab'=>'configuration'), admin_url('options-general.php') ));

	}

	/**
	 * This class has it's own settings page, to ensure it can always be called
	 *
	 * @return bool
	 */
	public function is_settings_page()
	{
		if (isset($_GET["page"]) && ($_GET["page"] == "rlrsssl_really_simple_ssl" || $_GET["page"] == "really-simple-ssl") ) {
			return true;
		}
		return false;
	}

	/**
	 * Get CSP rules for any type or output type
	 * @param false  $html_output
	 *
	 * @return string
	 */
	public function get_csp_rules( $html_output = false ) {
		//script-src-elem 'self' 'unsafe-inline' https://goingtoamerica.nl http://pvcsd.org; style-src 'self' https://fonts.googleapis.com 'unsafe-inline';
		$rules = array();
		global $wpdb;
		//The base content security policy rules, used in later functions to generate the Content Security Policy
		$rules['default-src'] = "default-src 'self' ;";
		$rules['script-src'] = "script-src 'self' 'unsafe-inline' ;";
		$rules['script-src-elem'] = "script-src-elem 'self' 'unsafe-inline' ;";
		$rules['style-src'] = "style-src 'self' 'unsafe-inline' ;";
		$rules['style-src-elem'] = "style-src-elem 'self' 'unsafe-inline' ;";

		$table_name = $wpdb->base_prefix . "rsssl_csp_log";
		$rows = $wpdb->get_results("SELECT * FROM $table_name ORDER BY time DESC");
		if ( !empty($rows) ) {
			foreach ($rows as $row) {
				if ( $row->inpolicy === 'true' ) {
					$violatedirective = $row->violateddirective;
					$blockeduri = $row->blockeduri;
					//Get uri value
					$uri = rsssl_sanitize_uri_value($blockeduri);
					//Generate CSP rule based on input
					$rules = $this->generate_csp_rule($violatedirective, $uri, $rules);
				}
			}
		}
		$rules = implode(" ", $rules);

		if ( RSSSL_PRO()->rsssl_premium_options->get_networkwide_option('rsssl_content_security_policy') === 'report-only' ) {
			$csp_violation_endpoint = get_rest_url(null, 'rsssl/v1/csp');
			if (RSSSL_PRO()->rsssl_premium_options->get_networkwide_option('rsssl_csp_report_token')) {
				$token = RSSSL_PRO()->rsssl_premium_options->get_networkwide_option('rsssl_csp_report_token');
			} else {
				$token = time();
				RSSSL_PRO()->rsssl_premium_options->update_networkwide_option('rsssl_csp_report_token', $token);
			}
			//report-uri is deprecated, but still the most used in browsers. We simply add both
			$header = 'Report-To';
			$csp_endpoint_rules = "{'url': '".$csp_violation_endpoint."', 'group': 'csp-endpoint', 'max-age': 10886400}";
			$report_to_header = $this->wrap_header($header, $csp_endpoint_rules, $html_output);

			$header = 'Content-Security-Policy-Report-Only';
			$rules =  "$rules report-uri $csp_violation_endpoint?rsssl_apitoken=$token; report-to csp-endpoint";
			if ( RSSSL_PRO()->rsssl_premium_options->get_networkwide_option('rsssl_upgrade_insecure_requests') ) {
				$rules = "upgrade-insecure-requests; $rules";
			}
			$report_uri_header = $this->wrap_header($header, $rules, $html_output);

			return $report_to_header.$report_uri_header;

		} else if ( RSSSL_PRO()->rsssl_premium_options->get_networkwide_option('rsssl_content_security_policy') === 'enforce' ) {
			// If the upgrade-insecure-requests header has been enabled, add it to this CSP.
			$header = 'Content-Security-Policy';
			if ( RSSSL_PRO()->rsssl_premium_options->get_networkwide_option('rsssl_upgrade_insecure_requests') ) {
				$rules = "upgrade-insecure-requests; $rules";
			}
			return $this->wrap_header($header, $rules, $html_output);
		}
		return '';
	}

	/**
	 * Generate security headers, and insert in .htaccess file
	 * @param bool $force
	 */

	public function insert_security_headers( $force = false ) {
		$wpconfig_path = RSSSL()->really_simple_ssl->find_wp_config_path();
		$wpconfig = file_get_contents($wpconfig_path);
		if (
			( !$force && ! RSSSL_PRO()->rsssl_premium_options->is_settings_page() )
			|| rsssl_get_networkwide_option('rsssl_security_headers_method') !== 'advancedheaders'
			|| wp_doing_ajax()
			|| !RSSSL()->really_simple_ssl->ssl_enabled
			|| !current_user_can("manage_options")
			|| $this->is_header_check_running()
		) {
			return;
		}

		if ( defined('rsssl_pp_version') ) {
			return;
		}
		$dir = ABSPATH . 'wp-content';
		$file = $dir . '/advanced-headers.php';

		$rules = $this->generate_php_security_header_rules();
		$contents = '<?php' . "\n";
		$contents .= '/**' . "\n";
		$contents .= '* This file is created by Really Simple SSL pro' . "\n";
		$contents .= '* It is necessary to ensure all headers are set on all types of ' . "\n";
		$contents .= '* servers, hosting providers, and with most caching plugins as well' . "\n";
		$contents .= '*/' . "\n\n";
		$contents .= "defined('ABSPATH') or die();" . "\n\n";
		//allow disabling of headers for detection purposes
		$contents .= 'if ( isset($_GET["rsssl_header_test"]) && intval($_GET["rsssl_header_test"]) ===  '.$this->get_headers_nonce().' ) return;' . "\n\n";
		$contents .= $rules;

		// write to advanced-header.php file
		if ( is_writable( $dir ) ) {
			file_put_contents( $file, $contents );
		}

		if ( strpos($wpconfig, 'advanced-headers.php')===false && is_writable( $wpconfig_path ) ) {
			$rule = "if ( file_exists('" . $file . "') ) { " . "\n";
			$rule .= "\t" . "require_once '$file';" . "\n" . "}";
            $updated = preg_replace('/' . '<\?php' . '/', '<?php' . "\n" . $rule . "\n", $wpconfig, 1);
            file_put_contents($wpconfig_path, $updated);
		}


	}

	/**
	 * Generate and return a random nonce
	 * @return int
	 */
	public function get_headers_nonce(){
		if (!get_option("rsssl_header_detection_nonce")) {
			update_option("rsssl_header_detection_nonce", rand(1000, 999999999) , false);
		}
		return get_option("rsssl_header_detection_nonce");
	}

	/**
	 * @param bool $html_output
	 *
	 * @return array|string|string[]|null
	 *
	 * Get the PHP security headers rules. Use $html_output to get HTML output
	 */

	public function generate_php_security_header_rules( $html_output = false) {

		//Get values for each security header
		$hsts                          = RSSSL_PRO()->rsssl_premium_options->get_networkwide_option( 'rsssl_hsts' );
		$x_xss_protection              = RSSSL_PRO()->rsssl_premium_options->get_networkwide_option( 'rsssl_x_xss_protection' );
		$upgrade_insecure_requests     = RSSSL_PRO()->rsssl_premium_options->get_networkwide_option( 'rsssl_upgrade_insecure_requests' );
		$x_content_type_options        = RSSSL_PRO()->rsssl_premium_options->get_networkwide_option( 'rsssl_x_content_type_options' );
		$no_referrer_when_downgrade    = RSSSL_PRO()->rsssl_premium_options->get_networkwide_option( 'rsssl_no_referrer_when_downgrade' );
		$x_frame_options               = RSSSL_PRO()->rsssl_premium_options->get_networkwide_option( 'rsssl_x_frame_options' );
		$permissions_policy            = RSSSL_PRO()->rsssl_premium_options->get_networkwide_option( 'rsssl_turn_on_permissions_policy' );
		$block_third_party_popups      = RSSSL_PRO()->rsssl_premium_options->get_networkwide_option( 'rsssl_block_third_party_popups' );
		$share_resources_third_parties = RSSSL_PRO()->rsssl_premium_options->get_networkwide_option( 'rsssl_share_resources_third_parties' );

		$rule = '';
		if ( is_ssl() && $hsts ) {
			// Not adding  env=HTTPS causes errors on lots of servers.
			// Remove the HSTS header from the old block before adding it to the new block
			$hsts_preload = RSSSL_PRO()->rsssl_premium_options->get_networkwide_option("rsssl_hsts_preload");
			if ($hsts_preload ){
				$rule .= $this->wrap_header('Strict-Transport-Security', "max-age=63072000; includeSubDomains; preload", $html_output);
			} else {
				$rule .= $this->wrap_header('Strict-Transport-Security', "max-age=31536000", $html_output);
			}
		}

		// Do not add the upgrade-insecure-requests header here when CSP is enforced, CSP will include this option when it is enabled
		if (RSSSL_PRO()->rsssl_premium_options->get_networkwide_option('rsssl_content_security_policy') !== 'enforce' && RSSSL_PRO()->rsssl_premium_options->get_networkwide_option('rsssl_content_security_policy') !== 'report-only' ) {
			if ($upgrade_insecure_requests) {
				$rule .= $this->wrap_header('Content-Security-Policy', "upgrade-insecure-requests", $html_output);
			}
		} else if ( $html_output) {
			$rule .= $this->get_csp_rules( $html_output );
		} else {
			$rule .= $this->get_csp_rules();
		}

		if ( $x_xss_protection ) {
			$rule .= $this->wrap_header( 'X-XSS-Protection', "0", $html_output );
		}
		if ( $x_content_type_options ) {
			$rule .= $this->wrap_header( 'X-Content-Type-Options', "nosniff", $html_output );
		}
		if ( $no_referrer_when_downgrade ) {
			$rule .= $this->wrap_header( 'Referrer-Policy', "strict-origin-when-cross-origin", $html_output );
		}
		if ( $permissions_policy ) {
			$rule .= $this->generate_permissions_policy_header( $html_output );
		}
		if ( $x_frame_options ) {
			$rule .= $this->wrap_header( 'X-Frame-Options', "SAMEORIGIN", $html_output );
		}
		if ( $block_third_party_popups === 'no' ) {
			$rule .= $this->wrap_header( 'Cross-Origin-Opener-Policy', 'same-origin-allow-popups', $html_output );
		}
		if ( $block_third_party_popups === 'yes' ) {
			$rule .= $this->wrap_header( 'Cross-Origin-Opener-Policy', 'same-site', $html_output );
		}
		if ( $share_resources_third_parties === 'no' ) {
			$rule .= $this->wrap_header( 'Cross-Origin-Resource-Policy', 'same-origin', $html_output );
		}
		if ( $share_resources_third_parties === 'yes' ) {
			$rule .= $this->wrap_header( 'Cross-Origin-Resource-Policy', 'cross-origin', $html_output );
		}
		if ( $share_resources_third_parties === 'yes_own_domain' ) {
			$rule .= $this->wrap_header( 'Cross-Origin-Resource-Policy', 'same-site', $html_output );
		}

		if ( $html_output ) {
			// Remove last line break to end </code> on same line
			$rule = preg_replace('~<br>(?!.*<br>)~', '', $rule);
		}

		return $rule;
	}

	/**
	 * Get permissions policy rules
	 * @param bool $html_output
	 * @param string|bool $type
	 * @return string
	 */

	public function generate_permissions_policy_header( $html_output = false, $type = false) {
		$permissions_policy_values = RSSSL_PRO()->rsssl_premium_options->get_networkwide_option('rsssl_permissions_policy');
		$rules = '';
		foreach ( $permissions_policy_values as $policy => $value ) {
			switch ($value) {
				case '*':
					$rules .= $policy ."=(*)" .", ";
					break;
				case 'none':
					$rules .= $policy ."=()" .", ";
					break;
				case 'self':
					$rules .= $policy ."=(self)" .", ";
					break;
			}
		}

		// Remove last space and , from string
		$rules = substr_replace($rules ,"",-2);
		$php_rule = $this->wrap_header('Permissions-Policy', $rules);

		update_option('rsssl_pro_permissions_policy_headers_for_php', $php_rule);
		return $php_rule;
	}

	/**
	 * @param string $header
	 * @param string $rules
	 * @param bool $html_output
	 *
	 * @return string
	 */

	public function wrap_header( $header, $rules, $html_output = false ){
		if ($html_output) {
			$break = "<br>";
		} else {
			$break = "\n";
		}

		return 'header(' . '"' . $header . ": " . $rules . '"' . ');' . $break;
	}

	/**
	 * Generate CSP rules
	 *
	 * @param string $violateddirective
	 * @param string $uri
	 * @param array  $rules //previously detected rules
	 *
	 * @return array
	 */

	public function generate_csp_rule( $violateddirective, $uri, $rules )
	{
		// Check the violateddirective is valid
		if (isset($this->directives[$violateddirective])){
			// If the violated directive has an existing rule, update it
			if (isset($rules[$violateddirective])) {
				$rule_template = $this->directives[$violateddirective]; //'script-src-elem'   => "script-src-elem 'self' {uri}; ",

				//get existing rule
				$existing_rule = $rules[$violateddirective]; //'script-src-elem'   => "script-src-elem 'self' 'unsafe-inline';";
				//get part of directive before {uri}
				$rule_part = substr($rule_template, 0, strpos($rule_template, '{uri}')); //"script-src-elem 'self' "
				// URI can be both URL or a directive (for example script-src)
				// Check if the current rule already contains the URI
				if (strpos($existing_rule, $uri) !== false) {
					// If it contains the uri, do not add it again. Keep existing rule
					$new_rule = $existing_rule;
				} else {
					//does not contain the uri, add it.
					$new_rule = str_replace($rule_part, $rule_part . $uri . " ", $existing_rule);
				}
				//insert in array
				$rules[$violateddirective] = $new_rule;
			} else {
				$rules[$violateddirective] = str_replace('{uri}', $uri, $this->directives[$violateddirective]);
			}
		}

		return $rules;
	}

	/**
	 * Check if a header check is currently running
	 *
	 * @return bool
	 */
	public function is_header_check_running(){
//		@todo Header detection disabled, enable in future version
		return false;
		$headers = get_option('rsssl_class_headers_headers_check');
		if ( !$headers || !isset($headers['header_check_active']) ){
			return true;
		}

		return $headers['header_check_active'];
	}

	/**
	 * Get recommended security header function
	 * The result is stored in a transient, but this transient is cleared on each pageload when on the settings page.
	 *
	 * @param string $type
	 *
	 * @return mixed
	 *@todo Header detection disabled, enable in future version
	 */
//	public function get_detected_security_headers($type=false)
//	{
//		$nonce = $this->get_headers_nonce();
//		//if the settings were changed, we want to recheck the headers
//		if ( !$type ) {
//			delete_option( 'rsssl_class_headers_headers_check' );
//		}
//
//		if ( ! in_array($type, ['used', 'un_used']) ) {
//			$type = 'un_used';
//		}
//		$headers_method = rsssl_get_networkwide_option('rsssl_security_headers_method');
//		$check_headers = RSSSL_PRO()->rsssl_premium_options->security_headers;
//		$headers = get_option('rsssl_class_headers_headers_check');
//		if ( ! $headers ) {
//			//set a default
//			$headers = [
//				'curl_exists' => function_exists( 'curl_init' ),
//				'used' => [],
//				'un_used' => [],
//				'header_check_active' => true,
//			];
//			update_option( 'rsssl_class_headers_headers_check', $headers, false );
//
//			if ( function_exists( 'curl_init' ) ) {
//				$url     = get_site_url();
//				$ch               = curl_init();
//				$detected_headers_with_rsssl = [];
//				curl_setopt( $ch, CURLOPT_URL, $url );
//				curl_setopt( $ch, CURLOPT_RETURNTRANSFER, 1 );
//				curl_setopt( $ch, CURLOPT_TIMEOUT, 3 ); //timeout in seconds
//				curl_setopt( $ch, CURLOPT_HEADERFUNCTION,
//					function ( $curl, $header ) use ( &$detected_headers_with_rsssl ) {
//						$len    = strlen( $header );
//						$header = explode( ':', $header, 2 );
//						if ( count( $header ) < 2 ) // ignore invalid headers
//						{
//							return $len;
//						}
//						$detected_headers_with_rsssl[] = [
//							'name'  => strtolower( trim( $header[0] ) ),
//							'value' => trim( $header[1] ),
//						];
//
//						return $len;
//					}
//				);
//				curl_exec( $ch );
//				/**
//				 * also do a check without headers set by RSSSL
//				 */
//				if ( $headers_method === 'php' || $headers_method === 'advancedheaders' ) {
//					$url_exclude_rsssl     = get_site_url().'?rsssl_header_test='.$nonce;
//					$ch               = curl_init();
//					$detected_headers_without_rsssl = [];
//					curl_setopt( $ch, CURLOPT_URL, $url_exclude_rsssl );
//					curl_setopt( $ch, CURLOPT_RETURNTRANSFER, 1 );
//					curl_setopt( $ch, CURLOPT_TIMEOUT, 3 ); //timeout in seconds
//					curl_setopt( $ch, CURLOPT_HEADERFUNCTION,
//						function ( $curl, $header ) use ( &$detected_headers_without_rsssl ) {
//							$len    = strlen( $header );
//							$header = explode( ':', $header, 2 );
//							if ( count( $header ) < 2 ) // ignore invalid headers
//							{
//								return $len;
//							}
//							$detected_headers_without_rsssl[] = [
//								'name'  => strtolower( trim( $header[0] ) ),
//								'value' => trim( $header[1] ),
//							];
//
//							return $len;
//						}
//					);
//					curl_exec( $ch );
//					error_log("DETECTED HEADERS NO rsssl");
//					error_log(print_r($detected_headers_without_rsssl,true));
//				}
//
//				// Check if any headers have been found
//				error_log("DETECTED HEADERS with rsssl");
//				error_log(print_r($detected_headers_with_rsssl,true));
//
//				/**
//				 * Get used headers from list of headers. In case of php or advanced headers, we exclude RSSSL headers from this check.
//				 */
//				$detected_headers_for_used_detection = ($headers_method==='php' || $headers_method==='advancedheaders') ? $detected_headers_without_rsssl : $detected_headers_with_rsssl;
//				if ( ! empty( $detected_headers_for_used_detection ) ) {
//					// Loop through each header and check if it's one of the recommended security headers. If so, add to used_headers array.
//					foreach ( $detected_headers_for_used_detection as $detected_header ) {
//						$name  = $detected_header['name'];
//						$value = $detected_header['value'];
//						if ( ! is_string( $value ) ) {
//							continue;
//						}
//
//						foreach ( $check_headers as $header ) {
//							$name_pattern = $header['pattern']['header'];
//							$value_pattern = $header['pattern']['value'];
//							//the header pattern should always match. If there is a value pattern, it has to match as well.
//							if ( stripos( $name, $name_pattern ) !== false && ( !$value_pattern || stripos( $value, $value_pattern ) !== false) ) {
//								// Remove \n's
//								$value_sanitized                             = str_replace( PHP_EOL, '', $value );
//								$headers['used'][ $header['name'] ]['value'] = $value_sanitized;
//								if ( in_array( $header['name'], $headers['used'] ) ) {
//									$headers['used'][ $header['name'] ]['count'] = 2;
//								} else {
//									$headers['used'][ $header['name'] ]['count'] = 1;
//								}
//							}
//						}
//					}
//				}
//
//				/**
//				 * Get un-used headers from list of ALL headers, otherwise we get "not detected but enabled by RSSSL" notices
//				 */
//				$all_headers = $detected_headers_with_rsssl;
//				if ( ! empty( $all_headers ) ) {
//					$all_used_headers = [];
//					foreach ( $all_headers as $detected_header ) {
//						$name  = $detected_header['name'];
//						$value = $detected_header['value'];
//						if ( ! is_string( $value ) ) {
//							continue;
//						}
//
//						foreach ( $check_headers as $header ) {
//							$name_pattern = $header['pattern']['header'];
//							$value_pattern = $header['pattern']['value'];
//							//the header pattern should always match. If there is a value pattern, it has to match as well.
//							if ( stripos( $name, $name_pattern ) !== false && ( !$value_pattern || stripos( $value, $value_pattern ) !== false) ) {
//								// Remove \n's
//								$value_sanitized = str_replace( PHP_EOL, '', $value );
//								$all_used_headers[ $header['name'] ]['value'] = $value_sanitized;
//								if ( in_array( $header['name'], $all_used_headers ) ) {
//									$all_used_headers[ $header['name'] ]['count'] = 2;
//								} else {
//									$all_used_headers[ $header['name'] ]['count'] = 1;
//								}
//							}
//						}
//					}
//
//					// Now check which headers are unused. Compare the used headers against the $all_used_headers array.
//					foreach ( $check_headers as $header ) {
//						if ( ! array_key_exists( $header['name'], $all_used_headers ) ) {
//							// Do not add CORS headers to unused
//							if ( $header['name'] === 'Cross-Origin-Opener-Policy' || $header['name'] === 'Cross-Origin-Resource-Policy' ) continue;
//
//							// Header is unused, add to unused array
//							$headers['un_used'][] = $header['name'];
//						}
//					}
//				}
//			}
//			$headers['header_check_active'] = false;
//			update_option( 'rsssl_class_headers_headers_check', $headers, false );
//		}
//
//		if ( !$headers['curl_exists'] ) {
//			$headers = [
//				'curl_exists' => false,
//				'used' => [],
//				'un_used' => [],
//				'header_check_active' => true,
//			];
//			if (RSSSL()->rsssl_server->uses_htaccess() && file_exists(RSSSL()->really_simple_ssl->htaccess_file())) {
//				$htaccess = file_get_contents(RSSSL()->really_simple_ssl->htaccess_file());
//				foreach ($check_headers as $check_header){
//					if ( !preg_match("/".$check_header['pattern']."/", $htaccess, $check) ) {
//						$headers['un_used'][] = $check_header['name'];
//					} else {
//						$headers['used'][] = $check_header['name'];
//					}
//				}
//				$headers['header_check_active'] = false;
//				update_option( 'rsssl_class_headers_headers_check', $headers, false );
//			}
//		}
//		error_log("TOTAL DETECTED LIST ");
//		error_log(print_r($headers, true));
//		$headers = wp_parse_args( $headers, ['curl_exists'=>true, 'un_used'=>[], 'used'=>[] ]);		return $headers[$type];
//	}
}