<?php
defined('ABSPATH') or die("you do not have access to this page!");

use Elementor\Utils;

class rsssl_premium_options {
	private static $_this;
	public $has_http_redirect=false;
	public $permissions_policy_options;

	function __construct() {
		if ( isset( self::$_this ) ) {
			wp_die( sprintf( __( '%s is a singleton class and you cannot create a second instance.', 'really-simple-ssl-pro' ), get_class( $this ) ) );
		}

		self::$_this = $this;

        $this->permissions_policy_options = array(
			'accelerometer' => '*',
			'autoplay' => '*',
			'camera' => '*',
			'encrypted-media' => '*',
			'fullscreen' => '*',
			'geolocation' => '*',
			'microphone' => '*',
			'midi' => '*',
			'payment' => '*',
            'display-capture' => '*',
		);

        // Key must match rsssl_recommended_security_headers filter in free
        $this->security_headers = apply_filters( 'rsssl_pro_detected_security_headers', array(
	        [
		        'name' => 'Upgrade Insecure Requests',
		        'pattern' =>  ['header'=>'Content-Security-Policy', 'value' => 'upgrade-insecure-requests'],
		        'option_name' => 'rsssl_upgrade_insecure_requests',
		        'option_value' => 'upgrade-insecure-requests',
	        ],
	        [
		        'name' => 'Content Security Policy',
		        'pattern' =>  ['header'=>'Content-Security-Policy', 'value' => false],
		        'option_name' => 'rsssl_content_security_policy',
		        'option_value' => false,
	        ],
            [
                'name' => 'X-XSS protection',
                'pattern' =>  ['header'=>'X-XSS-Protection', 'value' => false],
                'option_name' => 'rsssl_x_xss_protection',
                'option_value' => '0',
            ],
            [
                'name' => 'X-Content Type Options',
                'pattern' =>  ['header'=>'X-Content-Type-Options', 'value' => false],
                'option_name' => 'rsssl_x_content_type_options',
                'option_value' => 'nosniff',
            ],
            [
                'name' => 'Referrer-Policy',
                'pattern' =>  ['header'=>'Referrer-Policy', 'value' => false],
                'option_name' => 'rsssl_no_referrer_when_downgrade',
                'option_value' => 'strict-origin-when-cross-origin',
            ],
            [
                'name' => 'Permissions-Policy',
                'pattern' =>  ['header'=>'Permissions-Policy', 'value' => false],
                'option_name' => 'rsssl_turn_on_permissions_policy',
                'option_value' => false,
            ],
            [
                'name' => 'X-Frame-Options',
                'pattern' =>  ['header'=>'X-Frame-Options', 'value' => false],
                'option_name' => 'rsssl_x_frame_options',
                'option_value' => 'SAMEORIGIN',
            ],
	        [
		        'name' => 'HTTP Strict Transport Security Preload',
		        'pattern' =>  ['header'=>'Strict-Transport-Security', 'value' => 'preload'],
		        'option_name' => 'rsssl_hsts_preload',
		        'option_value' => 'max-age=63072000; includeSubDomains; preload',
	        ],
	        [
		        'name' => 'HTTP Strict Transport Security',
		        'pattern' =>  ['header'=>'Strict-Transport-Security', 'value' => false],
		        'option_name' => 'rsssl_hsts',
		        'option_value' => array('max-age=31536000', 'max-age=63072000; includeSubDomains; preload'),
	        ]
        ));
        add_action( 'admin_enqueue_scripts', array( $this, 'enqueue_assets') );
		add_action( 'admin_init', array( $this, 'save_permissions_policy' ), 30 );
		add_action( 'admin_init', array( $this, 'auto_update_elementor_url'), 15);
        add_action( 'admin_init', array( $this, 'maybe_redirect_to_settings_page'), 40);
        add_action( 'activate_plugin', array( $this, 'delete_notice_cache'), 10, 3);
        add_action( 'admin_init', array( $this, 'check_upgrade' ), 10, 2 );

		add_action( "update_option", array( $this, "update_security_headers_method_functions" ), 10, 3 );
		add_action( "update_option_rlrsssl_options", array( $this, "maybe_clear_certificate_check_schedule" ), 30, 3 );
		add_action( "update_option_rsssl_content_security_policy", array( $this, "maybe_update_csp_activation_time" ), 20, 3 );

        // Security header option update hooks
		//@todo Header detection disabled, enable in future version
//		add_filter( "pre_update_option", array( $this, "check_if_option_disabled_by_user" ), 4, 3 );

		add_action( 'wp_loaded', array( $this, 'admin_mixed_content_fixer' ), 1 );
		add_action( 'admin_init', array( $this, 're_check_http_redirect' ), 2 );
		add_action( 'admin_init', array( $this, 'change_notices_free' ), 5 );
		add_action( "admin_init", array( $this, "insert_security_headers" ), 10 );
		add_action( 'admin_init', array($this, 'add_pro_settings'), 100 );
        add_action( 'admin_init', array($this, 'add_permissions_policy_settings'), 65 );

		add_action( "admin_notices", array($this, 'show_notice_csp_enabled_next_steps') );
		add_action( 'wp_ajax_dismiss_success_pro_multisite_notice', array($this,'dismiss_pro_multisite_notice_callback') );
		add_action( 'wp_ajax_dismiss_csp_next_steps_notice', array($this,'dismiss_csp_next_steps_notice_callback') );
		add_action( 'admin_print_footer_scripts', array($this, 'insert_csp_next_steps_dismiss') );
		$plugin = rsssl_pro_plugin;
		add_filter( "plugin_action_links_$plugin", array($this,'plugin_settings_link') );
		add_filter( "network_admin_plugin_action_links_$plugin", array($this,'plugin_settings_link') );

		add_filter( 'rsssl_progress_footer_right', array($this,'progress_footer_right') );
		add_filter( 'rsssl_progress_footer_left' , array($this,'progress_footer_left') );

		add_filter( 'rsssl_grid_tabs', array($this,'add_pro_tabs'),10, 1 );
		add_filter( 'rsssl_notices', array($this,'get_notices_list'),20, 1 );
		//@todo Header detection disabled, enable in future version
//        add_filter( 'rsssl_notices', array( $this, 'add_non_recommended_header_notices') );
		add_action( 'show_tab_security_headers', array($this, 'add_security_headers_page') );
		add_action( 'show_tab_premium', array($this, 'add_premium_page') );
		add_filter( 'rsssl_grid_items', array($this, 'add_pro_grid_items'));
		add_action( 'rsssl_system_status', array($this, 'add_pro_system_status'));
		add_action( 'rsssl_finished_text', array($this, 'finished_text'), 20 );
		add_action( 'rsssl_deactivate', array($this, 'deactivate'), 20 );

		//@todo Header detection disabled, enable in future version
//		add_action( 'admin_init', array($this, 'maybe_redirect_to_recheck_headers'), 999 );

		/**
		 *  WP Rocket
		 */
		if ( $this->get_networkwide_option('rsssl_security_headers_method') === 'advancedheaders' ){
			add_filter( 'rocket_htaccess_mod_rewrite', '__return_false' );
		} else if ($this->get_networkwide_option('rsssl_security_headers_method') === 'htaccess') {
			add_filter( 'before_rocket_htaccess_rules', array($this, 'add_rules_before_wp_rocket' ) );
		}

		add_filter( 'site_status_tests', array($this, 'override_site_health_headers' ), 100 );
    }

	static function this() {
		return self::$_this;
	}

	/**
     * As headers are already sent, removing them won't result in cleared headers. So we redirect to recheck.
     *
	 * @return void
	 */
    public function maybe_redirect_to_recheck_headers(){
	    if ( get_option('rsssl_recheck_security_headers' ) ) {
		    delete_option('rsssl_recheck_security_headers');
            $tab = isset($_GET['tab']) ? sanitize_title($_GET['tab']) : 'configuration';
		    wp_redirect(add_query_arg(array('page' => 'rlrsssl_really_simple_ssl', 'tab'=>$tab), admin_url('options-general.php') ));
            exit;
	    }
    }

	public function progress_footer_right( $html ) {
		return '';
	}

	/**
     * Footer for pro
	 * @param $html
	 *
	 * @return string
	 */
	public function progress_footer_left( $html ) {
        if ( RSSSL()->really_simple_ssl->ssl_enabled || !RSSSL()->really_simple_ssl->site_has_ssl)  {
			return '<span class="rsssl-footer-left">Really Simple SSL pro '.rsssl_pro_version.'</span>';
		} else {
            return '';
		}
    }

	/**
     * Drop de site health notices about headers, as we already have implemented them through pro
	 * @param array $tests
	 *
	 * @return array
	 */
	public function override_site_health_headers( $tests ) {
		unset($tests['direct']['rsssl-headers']);
    	return $tests;
	}

	/**
     * Override the finished text for free
	 * @return string
	 */
	public function finished_text(){
	    return __("SSL configuration finished!", "really-simple-ssl-pro");
    }

	/**
	 * Set some defaults, then redirect to settings page on activation.
	 */

    public function set_defaults(){
        if ( !current_user_can( 'manage_options' ) ) {
            return;
        }

		if ( !$this->get_networkwide_option('rsssl_pro_defaults_set') ) {
            //run this first
			//@todo Header detection disabled, enable in future version
			//$headers_set_by_thirdparty = $this->get_active_security_headers_set_by_thirdparty();

			// Only update default values for permissions policy when option hasn't been created
			if ( !$this->get_networkwide_option('rsssl_permissions_policy') ) {
				$this->update_networkwide_option('rsssl_permissions_policy', $this->permissions_policy_options);
			}

			$this->update_networkwide_option('rsssl_upgrade_insecure_requests', true );
			$this->update_networkwide_option('rsssl_x_xss_protection', true);
			$this->update_networkwide_option('rsssl_x_content_type_options', true);
			$this->update_networkwide_option('rsssl_no_referrer_when_downgrade', true);
			$this->update_networkwide_option('rsssl_content_security_policy', 'disabled');
			$this->update_networkwide_option('rsssl_block_third_party_popups', 'no');
			$this->update_networkwide_option('rsssl_share_resources_third_parties', 'yes');
            $this->update_networkwide_option('rsssl_security_headers_method', 'advancedheaders');
			update_option('rsssl_enable_csp_defaults', true, false);
			$this->set_permissions_policy_defaults();

			//upgrade settings to match possible third party headers, preventing duplicate headers
            //@todo Header detection disabled, enable in future version
//			foreach ( $headers_set_by_thirdparty as $header_name => $header_option ) {
//				$this->update_networkwide_option( $header_option, false );
//			}

			$this->insert_security_headers();
            $this->update_networkwide_option('rsssl_pro_defaults_set', true);
		}
	}

	/**
     * After the defaults have been set, the headers have been checked, verify if there are options set by third parties. If so, disable them.
	 * @return void
	 */
    public function after_defaults_set(){
	    if ( !$this->get_networkwide_option('rsssl_pro_defaults_set') && !$this->get_networkwide_option('rsssl_after_default_setup_completed') ) {

		    $this->update_networkwide_option('rsssl_after_default_setup_completed', true);
	    }
    }

	/**
	 *
	 */
	public function add_pro_system_status(){
		echo "TLS version up to date: " . $this->get_tls_version();
		if ($this->redirects_to_homepage()) {
			echo "Redirect to homepage detected \n";
		}
		if ($this->has_redirect_to_http()) {
			echo "Redirect to http:// detected \n";
		}
		if ($this->site_uses_cache()) {
			echo "Site uses caching \n";
		}
	}

	/**
	 * Add some css for the settings page
	 * @param string $hook
	 * @since  1.0
	 *
	 * @access public
	 *
	 */

	public function enqueue_assets($hook) {

		if ( $hook !== 'settings_page_really-simple-ssl' && $hook !== 'settings_page_rlrsssl_really_simple_ssl' ) return;

		//Datatables plugin to hide pagination when it isn't needed
		wp_register_script('rsssl-datatables-pagination',
			trailingslashit(rsssl_pro_url)
			. 'js/dataTables.conditionalPaging.min.js', array("jquery"), rsssl_pro_version);
		wp_enqueue_script('rsssl-datatables-pagination');

		if ( RSSSL_PRO()->rsssl_scan->has_cleared_scan_data() ) {
			$emptyscantable = __( 'You have scanned your site before, but the scan results are cleared from the cache. Run a new scan to see the results.', 'really-simple-ssl-pro' );
		} elseif ( RSSSL_PRO()->rsssl_scan->scan_completed_no_errors() ) {
			$emptyscantable = __("No mixed content has been found!", "really-simple-ssl-pro");
		} else {
			$emptyscantable = __("No scan done yet. Start a quick or full scan to identify any issues.", "really-simple-ssl-pro");
		}

		wp_register_style('rsssl-pro-datatables', rsssl_pro_url . 'css/datatables.min.css', "", rsssl_pro_version);
		wp_enqueue_style('rsssl-pro-datatables');
		wp_register_style('rsssl-pro-table-css', rsssl_pro_url . 'css/jquery-table.css', "", rsssl_pro_version);
		wp_enqueue_style('rsssl-pro-table-css');
		wp_enqueue_script('rsssl-pro-datatables', rsssl_pro_url . "js/datatables.min.js", array('jquery'), rsssl_pro_version, false);

		wp_register_style( 'rsssl-main', rsssl_pro_url . 'css/main.css',"", rsssl_pro_version );
		wp_enqueue_style( 'rsssl-main');

		wp_register_style( 'rsssl-pro-grid', rsssl_pro_url . 'grid/grid.min.css',"", rsssl_pro_version );
		wp_enqueue_style( 'rsssl-pro-grid');

		wp_enqueue_script('rsssl-main', rsssl_pro_url . 'js/rsssl.js', array('jquery'), rsssl_pro_version, true);
		wp_localize_script('rsssl-main','rsssl_ajax', array(
			'ajaxurl'=> admin_url( 'admin-ajax.php' ),
			'progress' => get_option('rsssl_progress', 0.1),
			'searchPlaceholder' => __("Search", "really-simple-ssl-pro"),
			'emptyScanTable' => $emptyscantable,
			'emptyCspTable' => __("No items reported yet. Come back at a later time.", "really-simple-ssl-pro"),
			'previous' => __("Previous", "really-simple-ssl-pro"),
			'next' => __("Next", "really-simple-ssl-pro"),
			'first' => __("First", "really-simple-ssl-pro"),
			'last' => __("Last", "really-simple-ssl-pro"),
		));
	}

	/**
	 * Run Upgrade procedure
	 */
	public function check_upgrade() {
        //wait with upgrading until header check completed
		if ( RSSSL_PRO()->rsssl_headers->is_header_check_running() ) {
			return;
		}

		if ( !current_user_can('manage_options' ) ) {
            return;
		}

        $prev_version = get_option( 'rsssl-pro-current-version', false );
		/**
		 * Set a "first version" variable, so we can check if some notices need to be shown
		 */
		if ( ! get_option( 'rsssl_first_version' ) ) {
			update_option( 'rsssl_first_version', rsssl_pro_version );
		}

		if ( rsssl_pro_version === $prev_version ) {
            return;
		}

		if ( $prev_version && version_compare( $prev_version, '2.1.19', '<' ) ) {
			do {
				if ( ! file_exists( RSSSL()->really_simple_ssl->htaccess_file() ) ) {
					break;
				}

				if ( RSSSL()->really_simple_ssl->do_not_edit_htaccess ) {
					break;
				}

				$htaccess = file_get_contents( RSSSL()->really_simple_ssl->htaccess_file() );
				if ( ! is_writable( RSSSL()->really_simple_ssl->htaccess_file() ) ) {
					break;
				}

				$htaccess = preg_replace( "/#\s?BEGIN\s?Really_Simple_SSL_HSTS.*?#\s?END\s?Really_Simple_SSL_HSTS/s",
					"", $htaccess );
				$htaccess = preg_replace( "/\n+/", "\n", $htaccess );

				// Save changes
				file_put_contents( RSSSL()->really_simple_ssl->htaccess_file(), $htaccess );

				// Re-run insertion of security headers to make sure HSTS is inserted.
				$this->insert_security_headers(true );

			} while (0);

			// Upgrade elementor option to prefixed one, also since 2.1.19
			update_option( 'rsssl_elementor_upgraded', get_option( 'elementor_upgraded' ) );
			delete_option( 'elementor_upgraded' );
		}

		if ( $prev_version && version_compare( $prev_version, '4.0.0', '<' ) ) {
			if (RSSSL()->really_simple_ssl->hsts) {
				update_option('rsssl_hsts', true);
			}
		}

		if ( $prev_version && version_compare( $prev_version, '4.0.5', '<' ) ) {
			if (
				RSSSL()->rsssl_server->get_server() === 'nginx'
				&& RSSSL_PRO()->rsssl_premium_options->security_header_enabled()
				|| function_exists('is_wpe') && is_wpe()
			) {
				$this->update_networkwide_option('rsssl_security_headers_method', 'advancedheaders');
			} else {
				$this->update_networkwide_option('rsssl_security_headers_method', 'htaccess');
			}
        }

        if ( $prev_version && version_compare( $prev_version, '4.1', '<' ) ) {
	        // Upgrade Feature Policy to Permissions Policy
	        if ( $this->get_networkwide_option( 'rsssl_turn_on_feature_policy ' ) ) {
		        $this->update_networkwide_option( 'rsssl_turn_on_permissions_policy', true );
		        $this->delete_networkwide_option( 'rsssl_turn_on_feature_policy' );
	        }

	        if ( $this->get_networkwide_option( 'rsssl_feature_policy' ) ) {
		        $values = $this->get_networkwide_option( 'rsssl_feature_policy' );
		        $this->update_networkwide_option( 'rsssl_permissions_policy', $values );
		        $this->delete_networkwide_option( 'rsssl_feature_policy' );
	        }

	        if ( $this->get_networkwide_option( 'rsssl_pro_feature_policy_headers_for_php ' ) ) {
		        $this->update_networkwide_option( 'rsssl_pro_permissions_policy_headers_for_php', true );
		        $this->delete_networkwide_option( 'rsssl_pro_feature_policy_headers_for_php' );
	        }

	        if ( $this->get_networkwide_option( 'rsssl_turn_on_permissions_policy ' ) ) {
		        $this->replace_feature_policy_rules();
	        }
        }

        if ( $prev_version && version_compare( $prev_version, '4.1.3', '<' ) ) {
            $upgrade_insecure_requests = $this->get_networkwide_option('rsssl_content_security_policy');
            if ($upgrade_insecure_requests) {
                $this->update_networkwide_option('rsssl_upgrade_insecure_requests', $upgrade_insecure_requests);
            }
            $this->delete_networkwide_option('rsssl_content_security_policy');

            if ( $this->get_networkwide_option('rsssl_enable_csp_reporting') ) {
                $this->update_networkwide_option('rsssl_content_security_policy', 'report-only');
            }

            if ( $this->get_networkwide_option('rsssl_add_csp_rules_to_htaccess') ) {
                $this->update_networkwide_option('rsssl_content_security_policy', 'enforce');
            }

            $this->delete_networkwide_option('rsssl_enable_csp_reporting');
            $this->delete_networkwide_option('rsssl_add_csp_rules_to_htaccess');

            $this->delete_networkwide_option('rsssl_csp_reporting_dismissed_timestamp');
            $this->delete_networkwide_option('rsssl_pro_csp_notice_next_steps_notice_postponed');

            // Defaults have been set before
            $this->update_networkwide_option('rsssl_pro_defaults_set', true );
        }

        if ( $prev_version && version_compare( $prev_version, '4.1.7', '<' ) ) {
            if (function_exists('is_wpe') && is_wpe()) {
                $this->update_networkwide_option('rsssl_security_headers_method', 'php');
            }
        }

		if ( $prev_version && version_compare( $prev_version, '4.1.9', '<' ) ) {
			$this->insert_security_headers(true );
		}
		
		if ( $prev_version && version_compare( $prev_version, '5.2', '<' ) ) {
		    if ( !$this->get_networkwide_option('rsssl_security_headers_method') && $this->get_networkwide_option('rsssl_enable_php_headers') !== false) {
			    if ( $this->get_networkwide_option( 'rsssl_enable_php_headers' ) ) {
				    $this->update_networkwide_option( 'rsssl_security_headers_method', 'php' );
			    } else {
				    $this->update_networkwide_option( 'rsssl_security_headers_method', 'htaccess' );
			    }
		    }

		    $this->delete_networkwide_option('rsssl_enable_php_headers');
		}

        if ( $prev_version && version_compare( $prev_version, '5.3.1', '<' ) ) {

            if ( $this->get_networkwide_option('rsssl_cross_origin_policies' ) === 'same-origin' ) {
                $this->update_networkwide_option('rsssl_block_third_party_popups', 'yes' );
                $this->update_networkwide_option('rsssl_share_resources_third_parties', 'no' );
            }

            if ( $this->get_networkwide_option('rsssl_cross_origin_policies') === 'cross-origin') {
                $this->update_networkwide_option('rsssl_block_third_party_popups', 'no' );
                $this->update_networkwide_option('rsssl_share_resources_third_parties', 'yes' );
            }

            if ( $this->get_networkwide_option('rsssl_cross_origin_policies') === 'disabled' ) {
		        $this->update_networkwide_option('rsssl_block_third_party_popups', 'no' );
		        $this->update_networkwide_option('rsssl_share_resources_third_parties', 'yes' );
	        }

            $this->delete_networkwide_option('rsssl_cross_origin_policies' );
            $this->insert_security_headers(true );
        }

		if ( $prev_version && version_compare( $prev_version, '5.4.0', '<' ) ) {
			$this->update_permissions_policy_features();
			$this->delete_networkwide_option( 'rsssl_expect_ct' );
			$this->insert_security_headers( true );
		}

        if ( $prev_version && version_compare( $prev_version, '5.5.1', '<' ) ) {
            if ( $this->get_networkwide_option('rsssl_security_headers_method') === 'php' ) {
	            $this->update_networkwide_option( 'rsssl_security_headers_method', 'advancedheaders' );
                RSSSL_PRO()->rsssl_headers->insert_security_headers( true );
            }
		}

		//@todo Header detection disabled, enable in future version
//		if ( $prev_version && version_compare( $prev_version, '5.5.0', '<' ) ) {
//            //upgrade settings to match possible third party headers, preventing duplicate headers
//
//			$headers_set_by_thirdparty = $this->get_active_security_headers_set_by_thirdparty();
//            foreach ($headers_set_by_thirdparty as $header_name => $header_option ){
//                $this->update_networkwide_option($header_option, false);
//            }
//			$this->insert_security_headers(true );
//		}

		update_option( 'rsssl-pro-current-version', rsssl_pro_version );
	}

    /**
     * Maybe redirect to settings page
     * @since 4.1.4
     */

	public function maybe_redirect_to_settings_page() {
        if ( get_transient('rsssl_pro_redirect_to_settings_page' ) ) {
            delete_transient('rsssl_pro_redirect_to_settings_page' );
            if ( !$this->is_settings_page() ) {
			    if ( is_multisite() && is_super_admin() ) {
				    wp_redirect( add_query_arg(array('page' => 'really-simple-ssl'), network_admin_url('settings.php') )  );
				    exit;
                } else {
				    wp_redirect( add_query_arg(array('page'=>'rlrsssl_really_simple_ssl','tab'=>'configuration'), admin_url('options-general.php') ) );
				    exit;
                }
            }
        }
    }

	/**
	 * Add pro tab
	 */

	public function add_pro_tabs($tabs)
	{

		$tabs['premium'] = "Premium";
		return $tabs;
	}

	/**
	 * Add pro page
	 */

	public function add_premium_page() {
		if (!current_user_can('manage_options')) return;

		RSSSL()->really_simple_ssl->render_grid( $this->premium_grid() );
	}

	/**
	 * Replace the free support block with the support form and add the premium settings block
	 *
	 * @param array $items
     * @return array
	 */
	public function add_pro_grid_items($items) {
		// Replace free blocks with pro if Really Simple SSL Pro is active
		if (!current_user_can('manage_options')) return $items;

		$items['support'] = array(
			'title' => __("Premium support", "really-simple-ssl-pro"),
			'secondary_header_item' => '',
			'content' => rsssl_pro_template_path . 'support.php',
			'footer' => rsssl_pro_template_path . 'support-footer.php',
			'class' => 'half-height support-form',
			'type' => 'support',
			'can_hide' => true,
		);
		$items['plugins'] = array(
			'title' => __("Premium settings", "really-simple-ssl-pro"),
			'header' => rsssl_template_path . 'header.php',
			'content' => rsssl_pro_template_path . 'settings.php',
			'footer' => rsssl_pro_template_path.'premium-settings-footer.php',
			'type' => 'settings',
			'class' => 'rsssl-premium-settings half-height',
			'can_hide' => true,
            'instructions' => 'https://really-simple-ssl.com/how-to-set-security-headers-on-apache-and-nginx/'
		);

		return $items;
	}

	/**
	 * Add premium grid blocks
	 */
	public function premium_grid() {

		// Add hidden class for Permissions Policy and CSP if corresponding options are not enabled
		if (!$this->get_networkwide_option('rsssl_turn_on_permissions_policy')  || !$this->apply_networkwide_ssl_feature() ) {
			$permissions_policy_hidden = 'rsssl-hidden';
		} else {
			$permissions_policy_hidden = '';
		}

		if ($this->get_networkwide_option('rsssl_content_security_policy') == 'disabled'|| !$this->apply_networkwide_ssl_feature() ) {
			$csp_hidden = 'rsssl-hidden';
		} else {
			$csp_hidden = '';
		}

		do_action( "rsssl_modals" );

		$grid_items = array(
		    'scan' => array(
				'title' => __("Mixed content scan", "really-simple-ssl-pro"),
				'content' => rsssl_pro_template_path.'/scan.php',
				'footer' => rsssl_pro_template_path.'/scan-footer.php',
				'class' => 'regular rsssl-scan-container',
				'type' => 'scan',
				'instructions' => 'https://really-simple-ssl.com/knowledge-base/mixed-content-scan-overview/',
			),
			'security-headers' => array(
				'title' => "Security headers",
				'content' => rsssl_pro_template_path.'/security-headers.php',
				'footer' => rsssl_template_path.'/settings-footer.php',
				'class' => 'regular rsssl-security-headers',
				'type' => 'settings',
                'instructions' => 'https://really-simple-ssl.com/everything-you-need-to-know-about-security-headers/',
			),
			'permissions-policy' => array(
				'title' => "Permissions Policy",
				'content' => rsssl_pro_template_path.'/permissions-policy.php',
				'footer' => rsssl_template_path.'/settings-footer.php',
				'class' => "regular rsssl-datatables rsssl-permissions-policy $permissions_policy_hidden",
				'type' => 'settings',
                'instructions' => 'https://really-simple-ssl.com/knowledge-base/how-to-use-the-permissions-policy-header/',
			),
			'content-security' => array(
				'title' => __("Content Security Policy configuration", "really-simple-ssl-pro"),
				'header' => rsssl_pro_template_path.'/csp-header.php',
                'content' => rsssl_pro_template_path.'/csp.php',
				'class' => "regular rsssl-datatables  rsssl-content-security-policy $csp_hidden",
				'type' => 'all',
			),
		);

		return $grid_items;
	}

	/**
	 * If a user submits the re-check form, we run the check again.
	 */
	public function re_check_http_redirect() {
		if (!RSSSL()->really_simple_ssl->ssl_enabled) {
			$this->has_http_redirect = $this->has_redirect_to_http();
		} else {
			$this->has_http_redirect = false;
		}
	}

	/**
	 *  Run deactivation script
	 */
	public function deactivate(){
		if (!current_user_can('manage_options')) return;

		$this->remove_htaccess_rules('Really_Simple_SSL_SECURITY_HEADERS', true );

		if ($this->get_networkwide_option('rsssl_content_security_policy') === 'report-only' ) {
            $this->remove_htaccess_rules('Really_Simple_SSL_CSP_Report_Only', true);
        }

        if ($this->get_networkwide_option('rsssl_content_security_policy') === 'enforce' ) {
            $this->remove_htaccess_rules('Really_Simple_SSL_Content_Security_Policy', true );
        }

	}

	/**
	 * Change free notices
	 */

	public function change_notices_free(){

		remove_action('rsssl_activation_notice_inner', array(RSSSL()->really_simple_ssl, 'show_pro'), 40);

		if (!RSSSL()->really_simple_ssl->ssl_enabled && $this->has_http_redirect){
			remove_action('rsssl_activation_notice_inner', array(RSSSL()->really_simple_ssl, 'show_enable_ssl_button'), 50);
		}
		add_action('rsssl_activation_notice_inner' , array($this, 'show_scan_buttons_before_activation'), 40);
	}

	/**
	 * Run the replace url s function in Elementor to make sure all resources are loaded over https.
	 *
	 *@throws Exception
	 */

	public function auto_update_elementor_url() {

		if (!current_user_can('manage_options')) {
			return;
		}

		if ( !function_exists( 'rsssl_uses_elementor' ) ) {
			return;
		}

		if (defined('RSSSL_NO_ELEMENTOR_UPGRADE') && RSSSL_NO_ELEMENTOR_UPGRADE) {
			return;
		}

		if ( is_multisite() ) {
			if (!RSSSL()->rsssl_multisite->selected_networkwide_or_per_site ) {
				return;
			}

			if (!get_site_option('rsssl_ms_elementor_urls_upgraded')) {

				// Get sites chunked
				$nr_of_sites = 25;
				$current_public_offset = get_site_option('rsssl_ms_elementor_public_replace_progress');
				$current_private_offset = get_site_option('rsssl_ms_elementor_private_replace_progress');

				$args = array(
					'number'   => $nr_of_sites,
					'offset'   => $current_public_offset,
					'public'   => 1,
					'deleted'  => 0,
					'spam'     => 0,
					'archived' => 0,
				);
				$public_sites = get_sites( $args );

				$args = array(
					'number'   => $nr_of_sites,
					'offset'   => $current_private_offset,
					'public'   => 0,
					'deleted'  => 0,
					'spam'     => 0,
					'archived' => 0,
				);

				$private_sites = get_sites( $args );
				update_site_option('rsssl_ms_elementor_public_replace_progress', $current_public_offset+$nr_of_sites);
				update_site_option('rsssl_ms_elementor_private_replace_progress', $current_private_offset+$nr_of_sites);

				//set batch of sites
				if (count($public_sites) ==! 0) {
					foreach ( $public_sites as $site ) {
						$this->replace_elementor_url( $site );
					}
				}

				if (count($private_sites) ==! 0) {
					foreach ( $private_sites as $site ) {
						$this->replace_elementor_url( $site );
					}
				}

				if (count($public_sites) == 0 && count($private_sites) == 0) {
					update_site_option('rsssl_ms_elementor_urls_upgraded', true);
					update_site_option('rsssl_ms_elementor_public_replace_progress', 0);
					update_site_option('rsssl_ms_elementor_private_replace_progress', 0);
				}
			}

		}
		//run both for multisite as for single site. On ms, this ensures this function will run always when a specific site is loaded.
		$this->replace_elementor_url();


	}

	/**
	 * Replace URLs in elementor to https.
	 *
	 * @param  $site
	 *
	 * @throws Exception
	 */

	public function replace_elementor_url( $site=false ) {
		if (!current_user_can('manage_options')) {
			return;
		}

		if ($site) {switch_to_blog( $site->blog_id );}

		if ( function_exists( 'rsssl_uses_elementor' ) ) {
			if ( RSSSL()->really_simple_ssl->ssl_enabled && rsssl_uses_elementor() && !get_option( 'rsssl_elementor_upgraded' ) ) {

				$url  = home_url();
				$from = str_replace( 'https://', 'http://', $url );
				$to   = str_replace( 'http://', 'https://', $url );

				//non www
				$from_no_www = str_replace("http://www.", "http://", $from);
				$to_no_www = str_replace("https://www.", "https://", $to);

				//www
				$from_www = str_replace("http://", "http://www.", $from_no_www);
				$to_www = str_replace("https://", "https://www.", $to_no_www);

				try {
					if (class_exists('Elementor\Utils')) {
						Elementor\Utils::replace_urls( $from_no_www, $to_no_www );
						Elementor\Utils::replace_urls( $from_www, $to_www );
					}
				}
				catch(Exception $e) {
					error_log("replace URL from Elementor failed");
				}
				update_option( 'rsssl_elementor_upgraded', true );
			}
		}

		if ($site) {restore_current_blog();}
	}

	/**
	 * Activate the mixed content fixer on the admin when enabled.
	 */

	public function admin_mixed_content_fixer() {

		$admin_mixed_content_fixer = get_option("rsssl_admin_mixed_content_fixer");
		if (is_multisite() && RSSSL()->rsssl_multisite->mixed_content_admin) {
			$admin_mixed_content_fixer = true;
		}

		if (is_admin() && is_ssl() && $admin_mixed_content_fixer) {
			RSSSL()->rsssl_mixed_content_fixer->fix_mixed_content();
		}
	}

	/**
	 * Validate options on save
	 *
	 * @param $input
	 *
	 * @return int|string
	 */

	public function options_validate($input){
		if ($input==1){
			$validated_input = 1;
		}else{
			$validated_input = "";
		}
		return $validated_input;
	}

	/**
	 * Validate a text field
	 */

	public function options_validate_text($input)
	{
		if (!current_user_can('manage_options')) {return '';}
		return sanitize_text_field($input);
	}


	/**
	 *
	 * Checks if a redirect to http:// is active to prevent redirect loop issues
	 * Since 2.0.20
	 *
	 * @access public
	 *
	 */

	public function has_redirect_to_http()
	{
		//run this function only once
		$detected_redirect = get_option('rsssl_redirect_to_http_check');
		$force_check = false;

		//but if the user explicitly rechecks, run it again.
		if (isset($_POST['rsssl-check-redirect'])) $force_check = true;

		if ($force_check || !$detected_redirect){
			//make sure this redirect check only happens once by immediately setting a value
			update_option('rsssl_redirect_to_http_check', 'https');
			$url = site_url();
			if (!function_exists('curl_init')) {
				return false;
			}

			//CURLOPT_FOLLOWLOCATION might cause issues on php<5.4
			if (version_compare(PHP_VERSION, '5.4') < 0) {
				return false;
			}

			//Change the http:// domain to https:// to test for a possible redirect back to http://.
			$url = str_replace("http://", "https://", $url);
			//Follow the entire redirect chain.
			$ch = curl_init($url);
			curl_setopt($ch, CURLOPT_NOBODY, 1);
			curl_setopt($ch, CURLOPT_FOLLOWLOCATION, 1); // follow redirects
			curl_setopt($ch, CURLOPT_AUTOREFERER, 1); // set referer on redirect
			curl_setopt($ch, CURLOPT_CONNECTTIMEOUT, 3);
			curl_setopt($ch, CURLOPT_TIMEOUT, 3); //timeout in seconds
			curl_exec($ch);
			//$target is the endpoint of the redirect chain
			$target = curl_getinfo($ch, CURLINFO_EFFECTIVE_URL);
			curl_close($ch);

			//Check for http:// needle in target
			$http_needle = 'http://';

			$pos = strpos($target, $http_needle);

			if ($pos !== false) {
				//There is a redirect back to HTTP.
				$detected_redirect = 'http';
			} else {
				$detected_redirect = 'https';
			}
			update_option('rsssl_redirect_to_http_check', $detected_redirect);
		}

		if ($detected_redirect === 'http') {
			return true;
		} else {
			return false;
		}
	}


	/**
	 * @return bool
	 *
	 * Detect if a redirect to homepage is active. Can cause issues with 404 images which are redirect to homepage, making it impossible to locate the origin
	 *
	 */

	public function redirects_to_homepage() {
		$redirect_checked = get_transient('rsssl_redirects_to_homepage');
		if (!$redirect_checked) {
			$redirect_checked = 'OK';
			try {
				if ( ini_get('allow_url_fopen') ) {
					$non_existing_page = str_replace('http://', 'https://', site_url() . "/really-simple-ssl-404-test");

					stream_context_set_default( array(
						'ssl' => array(
							'verify_peer' => false,
							'verify_peer_name' => false,
						),
					));

					$http_headers  = @get_headers( $non_existing_page );
					if ( $http_headers && isset( $http_headers[0]) ) {
						$response_code = substr( $http_headers[0], 9, 3 );

						if ( $response_code === '301' || $response_code === '302' ) {
							//301/302 detected, check if destination matches the site URL. If so, we have a redirect to homepage
							foreach ($http_headers as $key => $header ){
								if (stripos($header, 'location') !== false ) {
									if ( preg_match( '/(http:\/\/|https:\/\/|\/\/)([\w.,;@?^=%&:()\/~+#!\-*].*)/i', $header, $matches )
									) {
										$location = $matches[0];
										//should contain http (end point is http) AND match either with http or https site_url.
										if ( strpos($location, 'http://' )!==false && (str_replace('http://', 'https://', $location) === site_url() ||  $location === site_url()) ) {
											$redirect_checked = 'REDIRECTING';
										}
									}

									break;
								}
							}
						}
					}

				}
			} catch (Exception $e) {
				$redirect_checked = 'OK';
			}
			$expiration = HOUR_IN_SECONDS;
			if ( $redirect_checked === 'OK' ) {
				$expiration = YEAR_IN_SECONDS;
			}
			set_transient('rsssl_redirects_to_homepage', $redirect_checked, $expiration );
		}

		if ( $redirect_checked === 'OK' ) {
			return false;
		} else {
			return true;
		}

	}

	/**
	 * Get the TLS version. Default to version 1.2 if cURL cannot complete.
	 * @return string
	 *
	 *
	 *
	 */

	public function get_tls_version() {

		$tls_version = 'not-found';
		if ( function_exists( 'curl_init' ) ) {
			$tls_version = get_transient('rsssl_tls_version');

			$possible_outcomes = array(
				'outdated',
				'up-to-date',
				'not-found'
			);

			//upgrade to new values
			if (!in_array($tls_version, $possible_outcomes)) $tls_version = false;

			if (!$tls_version) {
				$tls_version = 'not-found';

				$ch = curl_init( 'https://www.howsmyssl.com/a/check' );
				curl_setopt( $ch, CURLOPT_RETURNTRANSFER, true );
				curl_setopt( $ch, CURLOPT_TIMEOUT, 3 ); //timeout in seconds
				$data = curl_exec( $ch );
				curl_close( $ch );
				$json = json_decode( $data );
				if (!empty($json->tls_version) ) {
					$tls_version = str_replace( "TLS ", "", $json->tls_version );
				}

				if ( $tls_version === '1.0' || $tls_version === '1.1' ) {
					$tls_version = 'outdated';
				} else {
					$tls_version = 'up-to-date';
				}

				set_transient('rsssl_tls_version', $tls_version, MONTH_IN_SECONDS);
			}
		}
		return $tls_version;
	}

    /**
     * @param $oldvalue
     * @param $newvalue
     * @param $option
     * @return void
     *
     * Maybe update CSP activation time
     */
	public function maybe_update_csp_activation_time( $old_value, $value, $option) {
		if ( $old_value==$value ) {
			return;
		}

		if ( get_option("rsssl_csp_reporting_activation_time") ) {
            return;
        }

        update_option("rsssl_csp_reporting_activation_time", time());
	}

    /**
     * Remove advanced-headers.php if option is set to .htaccess
     * @since 5.5
     */

    public function update_security_headers_method_functions($option, $old_value, $value ) {
	    if ( $old_value === $value ) {
		    return;
	    }

	    if ( $option === 'rsssl_security_headers_method' ) {
            if ($old_value === 'nginxconf') {
	            set_transient( 'rsssl_show_nginxconf_notice', true, WEEK_IN_SECONDS );
            }
            if ( $value !== 'htaccess' ) {
                $this->remove_htaccess_rules( 'Really_Simple_SSL_SECURITY_HEADERS' );
                $this->remove_htaccess_rules( 'Really_Simple_SSL_CSP_Report_Only' );
                $this->remove_htaccess_rules( 'Really_Simple_SSL_Content_Security_Policy' );
            } else if ( $value !== 'advancedheaders' ) {
			    rsssl_pro_remove_advanced_headers();
            }

		    $this->maybe_flush_wprocket_htaccess();
		    update_option( 'rsssl_recheck_security_headers', true, false );
	    }
    }

	/**
     *
     * Check if one of the security headers option has been cosmetically enabled by RSSSL, and keep track if it was enabled by the user.
     *
	 * @param $value
	 * @param $old_value
	 * @param $option
	 *            //@todo Header detection disabled, enable in future version
	 * @return void
	 */
//	public function check_if_option_disabled_by_user( $value, $option, $old_value ) {
//        //if this is not posted by the user (save action) we skip
//        if ( !isset($_POST['option_page']) ) {
//            return $value;
//        }
//
//        $option_names = array_column($this->security_headers, 'option_name');
//        if ( in_array($option, $option_names) ) {
//            if ( $option==='rsssl_content_security_policy') {
//                return $value;
//            }
//	        $headers_set_by_thirdparty = $this->get_active_security_headers_set_by_thirdparty();
//	        $disabled_by_user = get_option('rsssl_security_headers_set_by_host_disabled_by_user', [] );
//	        if ( in_array( $option, $disabled_by_user) ) {
//		        $key = array_search($option, $disabled_by_user);
//		        unset($disabled_by_user[$key]);
//	        }
//
//	        //this header is not set by really simple ssl.
//	        if ( in_array( $option, $headers_set_by_thirdparty )  ) {
//		        //this header is not set by really simple ssl, but user is trying to disable it. Give a notice about it
//		        if ( !$this->option_is_enabled($value) && !in_array( $option, $disabled_by_user) ) {
//			        $disabled_by_user[] = $option;
//		        }
//		        $value = false;
//            //xss is a special case, as it has to be set preferably by us, with the 0 option.
//	        } else if ( $option === 'rsssl_x_xss_protection') {
//                if (!$value ) {
//	                update_option('rsssl_xss_protection_disabled_by_user', true );
//                } else {
//	                delete_option('rsssl_xss_protection_disabled_by_user' );
//                }
//            }
//	        update_option('rsssl_security_headers_set_by_host_disabled_by_user', $disabled_by_user);
//        }
//
//        return $value;
//	}

    /**
     *
     * Delete notice transient after plugin activation
     * @since 4.1
     */

	public function delete_notice_cache( ) {
        delete_transient('rsssl_admin_notices');
    }

	/**
	 * Add settings fields
	 */

	public function add_pro_settings(){
		if (!class_exists('REALLY_SIMPLE_SSL')) {
            return;
        }

		if (!current_user_can('manage_options')) {
            return;
        }

		add_settings_section('rlrsssl_pro_settings_section', '', '', 'rlrsssl_pro_settings_page');

		$help_tip = RSSSL()->rsssl_help->get_help_tip( __("If your certificate expires, your site goes offline. Uptime robots don't alert you when this happens.", "really-simple-ssl-pro")." ".
		                                              sprintf(__("This option sends an email to the administrator email address (%s) when your certificate is about to expire within 2 weeks.", "really-simple-ssl-pro"), get_option('admin_email') ), $return=true );
		add_settings_field('id_cert_expiration_warning', $help_tip . __("Receive an email when your certificate is about to expire","really-simple-ssl-pro"), array($this,'get_option_cert_expiration_warning'), 'rlrsssl_pro_settings_page', 'rlrsssl_pro_settings_section');

		if ( !is_multisite() || is_multisite() && is_network_admin() || is_multisite() && !is_network_admin() && !RSSSL()->rsssl_multisite->mixed_content_admin ) {
			$help_tip = RSSSL()->rsssl_help->get_help_tip( __( "Use this option if you do not have the secure lock in the WordPress admin.", "really-simple-ssl-pro" ), $return = true );
		} elseif ( is_multisite() && !is_network_admin() && RSSSL()->rsssl_multisite->mixed_content_admin ) {
			$help_tip = RSSSL()->rsssl_help->get_help_tip( __( "This option is enabled on the network menu", "really-simple-ssl-pro" ), $return = true );
		}

		add_settings_field('id_admin_mixed_content_fixer', $help_tip . __("Mixed content fixer on the WordPress back-end","really-simple-ssl-pro"), array($this,'get_option_admin_mixed_content_fixer'), 'rlrsssl_pro_settings_page', 'rlrsssl_pro_settings_section');

		//add_settings_section('section_rssslpp', __("Pro", "really-simple-ssl-pro"), array($this, "section_text"), 'rlrsssl');
		register_setting( 'rlrsssl_pro_options', 'rsssl_admin_mixed_content_fixer', array($this,'options_validate') );
		register_setting( 'rlrsssl_pro_options', 'rsssl_cert_expiration_warning', array($this,'options_validate') );
		add_settings_section('rlrsssl_security_headers_section', '', '', 'rlrsssl_security_headers_page');

        $help_tip = RSSSL()->rsssl_help->get_help_tip(__("This is an additional header to force all incoming http:// requests to https://.", "really-simple-ssl-pro"), $return=true );
        add_settings_field('id_upgrade_insecure_requests', $help_tip . __("Upgrade insecure request", "really-simple-ssl-pro"), array($this, 'get_option_upgrade_insecure_requests'), 'rlrsssl_security_headers_page', 'rlrsssl_security_headers_section');
        register_setting('rlrsssl_security_headers', 'rsssl_upgrade_insecure_requests', array($this, 'options_validate'));
		$help_tip = RSSSL()->rsssl_help->get_help_tip(__("Enabling this setting will set the X-XSS-Protection header to 0 (disable XSS filter). Disabling this option will remove the header and leave your site vulnerable to XSS and sidechannel attacks.", "really-simple-ssl-pro"), $return=true );
		add_settings_field('id_x_xss_protection', $help_tip . "Cross-site scripting (X-XSS) protection", array($this, 'get_option_x_xss_protection'), 'rlrsssl_security_headers_page', 'rlrsssl_security_headers_section');
		register_setting('rlrsssl_security_headers', 'rsssl_x_xss_protection', array($this, 'options_validate'));
		$help_tip = RSSSL()->rsssl_help->get_help_tip(__("This header prevents MIME-sniffing, which can be used to upload malicious files disguised as another content type.", "really-simple-ssl-pro"), $return=true );
		add_settings_field('id_x_content_type_options', $help_tip . "X Content Type Options", array($this, 'get_option_x_content_type_options'), 'rlrsssl_security_headers_page', 'rlrsssl_security_headers_section');
		register_setting('rlrsssl_security_headers', 'rsssl_x_content_type_options', array($this, 'options_validate'));
		$help_tip = RSSSL()->rsssl_help->get_help_tip(__("This header limits referrer information included in requests to other websites to the domain name only.", "really-simple-ssl-pro"), $return=true );
		add_settings_field('id_no_referrer_when_downgrade', $help_tip . "Referrer Policy", array($this, 'get_option_no_referrer_when_downgrade'), 'rlrsssl_security_headers_page', 'rlrsssl_security_headers_section');
		register_setting('rlrsssl_security_headers', 'rsssl_no_referrer_when_downgrade', array($this, 'options_validate'));
		$help_tip = RSSSL()->rsssl_help->get_help_tip(__("This header prevents your site from being loaded in an iFrame on other domains. This is used to prevent clickjacking attacks.", "really-simple-ssl-pro"), $return=true );
		add_settings_field('id_x-frame-options', $help_tip . "X Frame Options", array($this, 'get_option_x_frame_options'), 'rlrsssl_security_headers_page', 'rlrsssl_security_headers_section');
		register_setting('rlrsssl_security_headers', 'rsssl_x_frame_options', array($this, 'options_validate'));
		
		$help_tip = RSSSL()->rsssl_help->get_help_tip(__("This header allows you to restrict browser features on your own and embedded pages.", "really-simple-ssl-pro")." ".
                                                      __("It is recommended to enable this feature as soon as your site is running smoothly on SSL, as it greatly improves security.", "really-simple-ssl-pro"), $return=true );
		add_settings_field('id_turn_on_permissions_policy',  $help_tip . "Permissions Policy", array($this,'get_option_turn_on_permissions_policy'), 'rlrsssl_security_headers_page', 'rlrsssl_security_headers_section');
		register_setting('rlrsssl_security_headers', 'rsssl_turn_on_permissions_policy', array($this, 'options_validate'));

		$help_tip = RSSSL()->rsssl_help->get_help_tip(__("The Content Security Policy allows you to specify which resources to allow on your site.", "really-simple-ssl-pro"), $return=true );
		add_settings_field('id_content_security_policy', $help_tip . __("Content Security Policy","really-simple-ssl-pro"), array($this,'get_option_content_security_policy'), 'rlrsssl_security_headers_page', 'rlrsssl_security_headers_section');
		register_setting( 'rlrsssl_security_headers', 'rsssl_content_security_policy', '' );

		$help_tip = RSSSL()->rsssl_help->get_help_tip(__("HSTS, HTTP Strict Transport Security improves your security by forcing all your visitors to go to the SSL version of your website for at least a year.", "really-simple-ssl-pro")." ".__("It is recommended to enable this feature as soon as your site is running smoothly on SSL, as it improves your security.", "really-simple-ssl-pro"), $return=true );
		add_settings_field('id_hsts', $help_tip . "Strict Transport Security (HSTS)", array($this,'get_option_hsts'), 'rlrsssl_security_headers_page', 'rlrsssl_security_headers_section');
		register_setting( 'rlrsssl_security_headers', 'rsssl_hsts', array($this,'options_validate') );

		$help_tip = RSSSL()->rsssl_help->get_help_tip(__("The preload list offers even more security, as browsers will be instructed to load the site over HTTPS, even before the first visit.", "really-simple-ssl-pro")." ".
		                                              __("Please ensure that your site is running over SSL smoothly, and note that all subdomains, including www and non-www domain need to be HTTPS!", "really-simple-ssl-pro"), $return=true );
		add_settings_field('id_hsts_preload', $help_tip . __("Configure your site for the HSTS preload list","really-simple-ssl-pro"), array($this,'get_option_hsts_preload'), 'rlrsssl_security_headers_page', 'rlrsssl_security_headers_section');
		register_setting( 'rlrsssl_security_headers', 'rsssl_hsts_preload', array($this,'options_validate') );

		$help_tip = RSSSL()->rsssl_help->get_help_tip(__("Select if your site uses popups. Used to generate the Cross-Origin policy security header.", "really-simple-ssl-pro"), $return=true );
		add_settings_field('id_block_third_party_popups', $help_tip . __("Block third-party popups","really-simple-ssl-pro"), array($this,'get_option_block_third_party_popups'), 'rlrsssl_security_headers_page', 'rlrsssl_security_headers_section');
		register_setting( 'rlrsssl_security_headers', 'rsssl_block_third_party_popups', '' );

        $help_tip = RSSSL()->rsssl_help->get_help_tip(__("Select if your site is used as a resource on other websites. Used to generate the Cross-Origin policy security header.", "really-simple-ssl-pro"), $return=true );
        add_settings_field('id_share_resources_third_parties', $help_tip . __("Sharing resources with third parties","really-simple-ssl-pro"), array($this,'get_option_share_resources_third_parties'), 'rlrsssl_security_headers_page', 'rlrsssl_security_headers_section');
        register_setting( 'rlrsssl_security_headers', 'rsssl_share_resources_third_parties', '' );

        $help_tip = RSSSL()->rsssl_help->get_help_tip(__("Set the security headers with PHP. Should be enabled when using NGINX as the webserver, or when the htaccess security headers are not supported by your hosting company.", "really-simple-ssl-pro"), $return = true);
        add_settings_field('id_rsssl_security_headers_method', $help_tip . __("How to set the security headers", "really-simple-ssl-pro"), array($this, 'get_option_enable_php_headers'), 'rlrsssl_pro_settings_page', 'rlrsssl_pro_settings_section');
        register_setting('rlrsssl_pro_options', 'rsssl_security_headers_method', '');

		//add an overlay
		if ( !$this->apply_networkwide_ssl_feature() ) {
			$link_open = $link_close = '';
			if (is_super_admin()) {
				$link_open = '<a href="'.add_query_arg(array('page' => 'really-simple-ssl'), network_admin_url('settings.php') ).'">';
				$link_close = '</a>';
			}
			$string = sprintf(__("These settings are only available under the %snetwork settings%s.","really-simple-ssl-pro"), $link_open, $link_close);
			if (!defined('rsssl_pro_ms_template_path')) {
				$link_open = '<a href="https://really-simple-ssl.com/pro">';
				$link_close = '</a>';
				$string = sprintf(__("These settings are only available on the network settings, available with the %sAgency license%s.","really-simple-ssl-pro"), $link_open, $link_close);
            }
			add_settings_field('id_network_admin_only',  '<div class="rsssl-networksettings-overlay"><div class="rsssl-disabled-settings-overlay"><span class="rsssl-progress-status rsssl-open">multisite</span>'.
             $string
             .'</div></div>', array($this,'get_option_network_admin_only'), 'rlrsssl_security_headers_page', 'rlrsssl_security_headers_section');
		}
	}

    /**
     * Add Permissions Policy field
     */
    public function add_permissions_policy_settings() {
        if (!class_exists('REALLY_SIMPLE_SSL') && (!class_exists('REALLY_SIMPLE_SSL_PP'))) {return;}
        if (!current_user_can('manage_options')) {return;}
        add_settings_section('rlrsssl_permissions_policy_section', '', '', 'rlrsssl_permissions_policy_page');
        add_settings_field('id_permissions_policy', '', array($this, 'get_option_permissions_policy'), 'rlrsssl_permissions_policy_page', 'rlrsssl_permissions_policy_section');
        register_setting( 'rlrsssl_permissions_policy_group', 'rsssl_permissions_policy_option ' );
    }

	/**
     * Get a list of headers that were enabled, but not detected
	 * @return array
     *              //@todo Header detection disabled, enable in future version
	 */
//    public function undetected_enabled_headers(){
//        $security_headers = $this->security_headers;
//	    $set_but_not_detected = [];
//	    $unused_headers = RSSSL_PRO()->rsssl_headers->get_detected_security_headers('un_used');
//	    foreach ( $security_headers as $header ) {
//		    if ( in_array($header['name'], $unused_headers) && $this->get_networkwide_option( $header['option_name'] ) ){
//			    $set_but_not_detected[] = $header['name'];
//		    }
//	    }
//        return $set_but_not_detected;
//    }

	/**
     * Check the configuration for the headers
     * @todo Header detection disabled, enable in future version
	 * @return bool|string
	 */
//    public function all_enabled_headers_are_detected() {
//        if ( $this->get_networkwide_option('rsssl_security_headers_method') === 'php' || $this->get_networkwide_option('rsssl_security_headers_method') === 'advancedheaders' ) {
//	        $php_headers = true;
//        } else {
//            $php_headers = false;
//        }
//
//	    $set_but_not_detected = $this->undetected_enabled_headers();
//        //headers that were set, but not detected.
//        if ( count( $set_but_not_detected ) > 0 && $this->site_uses_cache() && $php_headers === false ) {
//            return 'enable_advanced_headers';
//        } else {
//            //al headers that are set, or we can't do anything about it, as php headers are already enabled, were detected.
//            return true;
//        }
//    }

	/**
	 * Get list of notices for the dashboard
     * @param array $notices
     *
     * @return array
	 */
	public function get_notices_list($notices)
	{
        unset($notices['pro_upsell']);
        unset($notices['recommended_security_headers_not_set']);
		//@todo Header detection disabled, enable in future version
//        $enabled_but_not_detected = $this->undetected_enabled_headers();
//		$enabled_but_not_detected = implode('<br>', $enabled_but_not_detected);
//		$enabled_but_not_detected = "<br><code style='padding: 0;'>" . $enabled_but_not_detected . "</code>";

//        $notices['recommended_security_headers_not_set'] = array(
//            'callback'  => 'RSSSL_PRO()->rsssl_premium_options->all_enabled_headers_are_detected',
//            'condition' => array( 'rsssl_ssl_enabled' ),
//            'score'     => 0,
//            'output'    => array(
//                'enable_advanced_headers'=> array(
//                    'msg'         => __( "Not all enabled security headers have been detected. Please enable the option 'advanced-headers.php' in the setting 'How to set the security headers'.", "really-simple-ssl-pro" )
//                                     . $enabled_but_not_detected,
//                    'icon' => 'open',
//                    'url' => 'https://really-simple-ssl.com/how-to-set-security-headers-on-apache-and-nginx/',
//                    'dismissible' => true
//                ),
//            ),
//        );

        if (get_transient('rsssl_show_nginxconf_notice')) {
	        $notices['switched_from_nginx_conf'] = array(
		        'callback'  => '_true_',
		        'condition' => array( 'rsssl_ssl_enabled' ),
		        'score'     => 0,
		        'output'    => array(
			        'true'=> array(
				        'msg'         => __( "You switched from the NGINX.conf setting to a new header setting. Don't forget to remove the NGINX.conf headers to prevent duplicate header issues.", "really-simple-ssl-pro" ),
				        'icon' => 'open',
				        'url' => 'https://really-simple-ssl.com/how-to-set-security-headers-on-apache-and-nginx/',
				        'dismissible' => true
			        ),
		        ),
	        );
        }


		//we remove some notices only for free:
		$notices['elementor'] = array(
			'callback' => 'rsssl_uses_elementor',
			'score' => 5,
			'output' => array(
				'uses-elementor' => array(
					'msg' => __("Elementor mixed content successfully converted.", "really-simple-ssl-pro"),
					'icon' => 'success',
					'dismissible' => true
				),
			),
		);

		$notices['redirect_loop_warning'] = array(
			'condition' => array('NOT rsssl_ssl_enabled'),
			'callback' => 'RSSSL_PRO()->rsssl_premium_options->has_redirect_to_http',
			'score' => 10,
			'output' => array(
				'false' => array(
					'msg' => __("No redirect to http detected.", "really-simple-ssl-pro"),
					'icon' => 'success'
				),
				'true' => array(
					'title' => __("Potential redirect loop.", "really-simple-ssl-pro"),
					'msg' => __("A redirect to http was detected. This might result in redirect loops.", "really-simple-ssl-pro").' '.sprintf(__("%sRead more%s.", "really-simple-ssl-pro"), '<a target="_blank" href="https://really-simple-ssl.com/knowledge-base/my-website-is-in-a-redirect-loop/">', '</a>')
                            .'<form action="" method="POST"><input type="submit" class="button" name="rsssl-check-redirect" value="'.__("Re-check the redirect","complianz").'"></form>',
					'icon' => 'warning',
					'admin_notice' => true,
				),
			),
		);

		$notices['pretty-permalinks-error'] = array(
            'condition' => array('rsssl_ssl_enabled'),
            'callback' => 'RSSSL_PRO()->rsssl_premium_options->pretty_permalinks_enabled_required',
            'score' => 10,
            'output' => array(
                'true' => array(
                    'title' => __("Pretty permalinks not enabled.", "really-simple-ssl-pro"),
                    'msg' => __("Pretty permalinks are not enabled on your site. This can cause issues with the REST API, used by Really Simple SSL Pro.", "really-simple-ssl-pro"),
                    'icon' => 'warning',
                    'url' => admin_url('options-permalink.php'),
                    'plusone' => true,
                    'dismissible' => false,
                ),
            ),
        );

		$notices['hsts_preload'] = array(
			'condition' => array('rsssl_ssl_enabled', 'RSSSL()->really_simple_ssl->contains_hsts'),
			'callback' => 'rsssl_pro_hsts_preload',
			'score' => 10,
			'output' => array(
				'true' => array(
					'msg' => sprintf(__("Your site has been configured for the HSTS preload list. If you have submitted your site, it will be preloaded. Click %shere%s to submit.", "really-simple-ssl-pro"),'<a target="_blank" href="https://hstspreload.org/?domain='.$this->non_www_domain().'">', '</a>' ),
					'icon' => 'success'
				),
				'false' => array(
					'url' => RSSSL()->really_simple_ssl->generate_enable_link($setting_name = 'hsts_preload', 'premium' ),
					'msg' => __("Your site is not yet configured for the HSTS preload list.", "really-simple-ssl-pro"),
					'icon' => 'open',
					'dismissible' => true,
				),
			),
		);

		$notices['tls_version'] = array(
			'condition' => array('rsssl_ssl_enabled'),
			'callback' => 'RSSSL_PRO()->rsssl_premium_options->get_tls_version',
			'score' => 10,
			'output' => array(
				'up-to-date' => array(
					'msg' => __("TLS version is up-to-date","really-simple-ssl-pro"),
					'icon' => 'success',
				),
				'outdated' => array(
                    'url' => 'https://really-simple-ssl.com/knowledge-base/deprecation-of-tls-1-0-and-1-1/',
					'msg' => __('Your site uses an outdated version of TLS. Upgrade to TLS 1.2 or TLS 1.3 to keep your site secure.',"really-simple-ssl-pro"),
					'icon' => 'warning',
					'dismissible' => true,
					'plusone' => true
				),
			),
		);

		$notices['redirect_to_homepage'] = array(
			'callback' => 'rsssl_redirect_to_homepage',
			'score' => 10,
			'output' => array(
				'redirect-to-homepage' => array(
                    'url' => "https://really-simple-ssl.com/knowledge-base/mixed-content-from-a-domain-image-source-caused-by-a-404-redirect-to-homepage/",
					'msg' => __('Your site redirects 404 pages to the http:// version of your homepage. This can cause mixed content issues with images.',"really-simple-ssl-pro"),
					'icon' => 'warning',
					'dismissible' => true,
					'plusone' => true
				),
			),
		);

        $notices['new_csp_entries'] = array(
            'condition' => array('rsssl_ssl_enabled'),
            'callback' => 'rsssl_pro_check_for_new_csp_entries',
            'score' => 10,
            'output' => array(
                'new-csp-rules' => array(
                    'msg' => __("You have new rules that can be added to your Content Security Policy.", "really-simple-ssl-pro"),
                    'icon' => 'open',
                    'plusone' => true,
                    'dismissible' => true
                ),
                'no-new-csp-rules' => array(
                    'msg' => __("No Content Security Policy violations found.", "really-simple-ssl-pro"),
                    'icon' => 'success'
                ),
                'report-paused' => array(
                    'msg' => __("Content Security Policy reporting is paused to limit server resources. Please start configuring your Content Security Policy and enable reporting again.", "really-simple-ssl-pro"),
                    'icon' => 'open',
                    'plusone' => true,
                ),
            ),
        );

        $notices['x_xss_protection_disabled'] = array(
            'callback' => 'rsssl_x_xss_protection_status',
            'score' => 10,
            'output' => array(
                'x-xss-disabled' => array(
                    'url' => RSSSL()->really_simple_ssl->generate_enable_link('rsssl_x_xss_protection', 'premium' ),
                    'msg' => __('You have disabled the X-XSS-Protection header. We recommend to enable the setting, this will set the X-XSS-Protection header to "0" (disable XSS filter). Disabling this option will remove the header and leave your site vulnerable to XSS and sidechannel attacks for users that use IE or Safari.',"really-simple-ssl-pro"),
                    'icon' => 'open',
                    'dismissible' => true,
                ),
            ),
        );
		//@todo Header detection disabled, enable in future version
//        $notices['security_headers_disabled_by_user_set_by_thirdparty'] = array(
//            'callback' => 'rsssl_security_headers_disabled_by_user_set_by_thirdparty_condition',
//            'score' => 10,
//            'output' => array(
//                'true' => array(
//                    'url' => "https://really-simple-ssl.com/how-to-find-where-security-headers-are-set/",
//                    'msg' => sprintf(__('You tried to disable %s, but this is not possible because it is set by a third party. Re-saving will dismiss the notice.',"really-simple-ssl-pro"), implode( ', ' , rsssl_security_headers_disabled_by_user_set_by_thirdparty() )),
//                    'icon' => 'open',
//                    'dismissible' => false,
//                ),
//            ),
//        );

		$start_scan = '';
		if ( !RSSSL_PRO()->rsssl_licensing->license_is_valid() ) {
			$start_scan = '&nbsp;<a href="'.add_query_arg(array('page'=>'rlrsssl_really_simple_ssl', 'tab' => 'premium', 'rsssl_start_scan' => 1 ), admin_url('options-general.php')).'">'.__("Start scan", "really-simple-ssl-pro").'</a>';
		}

        $activate_link = add_query_arg(
            array(
                'page' => 'rlrsssl_really_simple_ssl',
                'tab' => 'license'
            ), admin_url('options-general.php'));
        $activate =  ' '.sprintf(__("%sActivate%s your license.", "really-simple-ssl-pro"), '<a href="'.$activate_link.'">', '</a>');

        $notices['mixed_content_scan'] = array(
			'callback' => 'rsssl_pro_scan_notice',
			'score' => 10,
			'output' => array(
				'has-ssl-no-scan-errors' => array(
					'msg' => __("Great! Your scan last completed without errors.", "really-simple-ssl-pro"),
					'icon' => 'success'
				),
				'has-ssl-scan-has-errors' => array(
					'msg' => __("The last scan was completed with errors. Only migrate if you are sure the found errors are not a problem for your site.", "really-simple-ssl-pro"),
					'icon' => 'warning',
					'dismissible' => true
				),
				'no-ssl-no-scan-done-invalid-license' => array(
                    'msg' => __("You haven't scanned the site yet, you should scan your site to check for possible issues. Activate your license key to start a scan.", "really-simple-ssl-pro") . $activate,
					'icon' => 'open'
				),
                'no-scan-done' => array(
                    'msg' => __("You haven't scanned the site yet, you should scan your site to check for possible issues.", "really-simple-ssl-pro"). $start_scan,
                    'icon' => 'open'
                ),
				'no-ssl-no-scan-errors' => array(
					'msg' => __("Great! Your scan last completed without errors.", "really-simple-ssl-pro"),
					'icon' => 'success'
				),
				'no-ssl-scan-has-errors' => array(
					'msg' => __("The last scan was completed with errors. Are you sure these issues don't impact your site?", "really-simple-ssl-pro"),
					'icon' => 'warning',
					'dismissible' => true
				),
			),
		);

		$notices['ssl_enabled_networkwide'] = array(
			'callback' => 'rsssl_not_enabled_networkwide',
			'score' => 5,
			'output' => array(
				'htaccess' => array(
					'msg' => sprintf(__("You have a multisite environment. To leverage all configuration options in Really Simple SSL, we recommend to enabled the plugin networkwide, then use the 'activate per site' option to enable SSL per site.", "really-simple-ssl-pro"), '<a target="_blank" href="https://really-simple-ssl.com/pro#multisite">', '</a>'),
					'icon' => 'open',
					'dismissible' => 'true',
				),
			),
		);

		$link =  ' '.sprintf(__("You can upgrade on your %saccount%s.", "really-simple-ssl-pro"), '<a target="blank" href="https://really-simple-ssl.com/account">', '</a>');
        $notices['rsssl_pro_license_valid'] = array(
	        'condition' => array('NOT is_multisite'),//only show this one on single sites. Pro has it's own dashboard notice
	        'callback' => 'rsssl_pro_is_license_expired',
            'score' => 30,
            'output' => array(
                'expired' => array(
                    'title' => __("License", 'really-simple-ssl-pro'),
                    'msg' => __("Your Really Simple SSL Pro license key has expired. Please renew your license to continue receiving updates and premium support.", "really-simple-ssl-pro").$link,
                    'icon' => 'warning',
                    'plusone' => true,
                    'admin_notice' => false,
                ),
                'invalid' => array(
                    'title' => __("License", 'really-simple-ssl-pro'),
                    'msg' => __("Your Really Simple SSL Pro license key is not activated. Please activate your license to continue receiving updates and premium support.", "really-simple-ssl-pro").$activate,
                    'icon' => 'warning',
                    'plusone' => true,
	                'admin_notice' => false,
                ),
                'site_inactive' => array(
                    'title' => __("License", 'really-simple-ssl-pro'),
                    'msg' => __("This domain is not activated for this Really Simple SSL Pro license. Please activate the license for this domain.", "really-simple-ssl-pro").$link,
                    'icon' => 'warning',
                    'plusone' => true,
	                'admin_notice' => false,
                ),
                'no_activations_left' => array(
                    'title' => __("License", 'really-simple-ssl-pro'),
                    'msg' => __("You do not have any activations left on your Really Simple SSL Pro license. Please upgrade your plan for additional activations.", "really-simple-ssl-pro").$link,
                    'icon' => 'warning',
                    'plusone' => false,
	                'admin_notice' => false,
                ),
                'not-activated' => array(
	                'title' => __("License", 'really-simple-ssl-pro'),
	                'msg' => __("Your Really Simple SSL Pro license key hasn't been activated yet. You can activate your license key on the license tab.", "really-simple-ssl-pro").$activate,
                    'icon' => 'warning',
	                'plusone' => true,
	                'admin_notice' => false,
                ),
            ),
        );

        $notices['free_csp_compatibility'] = array(
            'condition' => array('rsssl_free_csp_incompatible' , 'rsssl_csp_enabled'),
            'callback' => '_true_',
            'score' => 5,
            'output' => array(
                'true' => array(
                    'msg' =>  __("Update Really Simple SSL Free to the latest version for optimal Content Security Policy compatibility.", "really-simple-ssl-pro"),
                    'icon' => 'open',
                    'dismissible' => 'true',
                    'plusone' => true,
                ),
            ),
        );

		$notices['free_compatibility'] = array(
			'condition' => array('rsssl_free_incompatible'),
			'callback' => '_true_',
			'score' => 5,
			'output' => array(
				'true' => array(
					'msg' =>  __("Update Really Simple SSL Free to the latest version for optimal compatibility.", "really-simple-ssl-pro"),
					'icon' => 'open',
					'dismissible' => 'true',
					'plusone' => true,
				),
			),
		);

        $notices['expect_ct_removed'] = array(
			'condition' => array('rsssl_upgraded_to_current_version'),
			'callback' => '_true_',
			'score' => 5,
			'output' => array(
				'true' => array(
                    'url' => 'https://really-simple-ssl.com/definition/what-is-expect-ct/',
					'msg' =>  __("The Expect-CT security header has been removed.", "really-simple-ssl-pro"),
					'icon' => 'open',
					'dismissible' => 'true',
					'plusone' => true,
				),
			),
		);
		$notices['xss-protection-updated'] = array(
			'condition' => array('rsssl_upgraded_to_current_version'),
			'callback' => '_true_',
			'score' => 5,
			'output' => array(
				'true' => array(
					'url' => 'https://really-simple-ssl.com/definition/what-is-xxss-protection/',
					'msg' =>  __("The X-XSS security header has been updated.", "really-simple-ssl-pro"),
					'icon' => 'open',
					'dismissible' => 'true',
					'plusone' => true,
				),
			),
		);

        $notices['wpcontent_not_writable'] = array(
	        'condition' => array('rsssl_ssl_enabled'),
			'callback' => 'rsssl_wp_content_not_writable',
			'score' => 5,
			'output' => array(
				'true' => array(
                    'url' => 'https://really-simple-ssl.com/htaccess-wp-config-file-not-writable/',
					'msg' =>  __("The wp-content folder is not writable. Set the wp-content folder permissions to writeable so Really Simple SSL can write the advanced-headers.php file.", "really-simple-ssl-pro"),
					'icon' => 'warning',
					'dismissible' => false,
					'plusone' => true,
				),
			),
		);

		$notices['security_headers_wpconfig_not_writable'] = array(
			'condition' => array('rsssl_ssl_enabled'),
			'callback' => 'rsssl_security_headers_wpconfig_not_writable',
			'score' => 10,
			'output' => array(
				'true' => array(
					'url' => 'https://really-simple-ssl.com/knowledge-base/htaccess-wp-config-file-not-writable/',
					'msg' => __("The wp-config file is not writable. Set the wp-config.php file permissions to writeable so Really Simple SSL can include the advanced-headers.php file.", "really-simple-ssl-pro"),
					'icon' => 'warning',
					'plusone' => true,
					'dismissible' => false,
				),
			),
		);

		return $notices;
	}

	/**
	 * Update option, network or single site
	 * @param string $name
	 * @param mixed $value
	 */
	public function update_networkwide_option($name, $value) {
		if (!current_user_can('manage_options')) return;
		if (is_multisite()) {
			update_site_option($name, $value);
		} else {
			update_option($name, $value);
		}
	}

	/**
	 * Get option, network or single site
	 * @param string name
	 * @return mixed
	 */
	public function get_networkwide_option($name) {
		if (!current_user_can('manage_options')) return;
		if (is_multisite()) {
			return get_site_option($name);
		} else {
			return get_option($name);
		}
	}

    /**
     * Delete option, network or single site
     * @param string name
     * @return mixed
     */
    public function delete_networkwide_option($name) {
        if (!current_user_can('manage_options')) return;
        if (is_multisite()) {
            return delete_site_option($name);
        } else {
            return delete_option($name);
        }
    }

	/**
	 * Insert option into settings form
	 * When multisite is enabled, these options are not accessible here
	 *
	 * @since  1.0.3
	 *
	 * @access public
	 *
	 */

	public function get_option_hsts() {
		$enabled = $this->get_networkwide_option('rsssl_hsts');
//		if ( $this->security_header_is_set_by_thirdparty('rsssl_hsts') ){
//            $confirm = "3d ($enabled)";
//			$enabled = true;
//		} else {
//			$confirm = "NOT 3d ($enabled)";
//		}
		?>
        <label class="rsssl-switch" id="rsssl-maybe-highlight-hsts-enabled">
            <input name="rsssl_hsts" size="40" value="1"
                   type="checkbox" <?php checked(1, $enabled, true) ?> />
            <span class="rsssl-slider rsssl-round"></span>

        </label>
		<?php
	}

    /**
     *
     * Upgrade insecure requests option
     *
     */

    public function get_option_upgrade_insecure_requests() {
	    $enabled = $this->get_networkwide_option('rsssl_upgrade_insecure_requests');
//        if ( $this->security_header_is_set_by_thirdparty('rsssl_upgrade_insecure_requests') ){
//	        $confirm = "3d ($enabled)";
//	        $enabled = true;
//        } else {
//	        $confirm = "NOT 3d ($enabled)";
//        }
	    ?>
        <label class="rsssl-switch">
            <input name="rsssl_upgrade_insecure_requests" size="40" value="1"
                   type="checkbox" <?php checked(1, $enabled, true) ?> />
            <span class="rsssl-slider rsssl-round"></span>
        </label>
        <?php
    }

	/**
	 * Insert option into settings form
	 * When multisite is enabled, these options are not accessible here
	 *
	 * @since  4.1.1
	 *
	 * @access public
	 *
	 */

	public function get_option_content_security_policy() {
		$csp = $this->get_networkwide_option('rsssl_content_security_policy');
//		if ( $this->security_header_is_set_by_thirdparty('rsssl_content_security_policy') ){
//			$confirm = "3d ($csp)";
//		} else {
//			$confirm = "NOT 3d ($csp)";
//		}
        $options = array(
            'disabled' => __('Disabled', 'really-simple-ssl-pro'),
            'report-only' => __('Reporting active', 'really-simple-ssl-pro'),
            'report-paused' => __('Reporting paused', 'really-simple-ssl-pro'),
            'enforce' => __('Enforce', 'really-simple-ssl-pro'),
        );
        ?>
        <select name="rsssl_content_security_policy">
            <?php foreach($options as $key => $name) {?>
            <option value=<?php echo $key?> <?php if ($csp == $key) echo "selected" ?>><?php echo $name ?>
                <?php }?>
        </select>
        <?php
        $comment = sprintf(__("This is an advanced feature. Only enable this if you know what you're doing. See %sour usage guide%s"), '<a href="https://really-simple-ssl.com/knowledge-base/how-to-use-the-content-security-policy-generator/">', '</a>');
        RSSSL()->rsssl_help->get_comment($comment);
    }

    /**
     * Select whether the site uses cross subdomain popups
     *
     * @since  5.2.4
     *
     * @access public
     *
     */

    public function get_option_block_third_party_popups() {

        $block_third_party_popups = $this->get_networkwide_option( 'rsssl_block_third_party_popups' );

        $options = array(
            'yes' => __('Yes', 'really-simple-ssl-pro'),
            'no' => __('No', 'really-simple-ssl-pro'),
        );

        ?>
        <select name="rsssl_block_third_party_popups">
            <?php foreach($options as $key => $name) {?>
            <option value=<?php echo $key?> <?php if ( $block_third_party_popups == $key ) echo "selected" ?>><?php echo $name ?>
                <?php } ?>
        </select>
        <?php
    }

    /**
     * Select whether the site is used as a resource for other sites
     *
     * @since  5.2.4
     *
     * @access public
     *
     */

    public function get_option_share_resources_third_parties() {

        $share_resources_third_parties = $this->get_networkwide_option( 'rsssl_share_resources_third_parties' );
        $options = array(
            'yes' => __('Sharing on', 'really-simple-ssl-pro'),
            'yes_own_domain' => __('Share only with own domain', 'really-simple-ssl-pro'),
            'no' => __('Sharing off', 'really-simple-ssl-pro'),
        );

        ?>
        <select name="rsssl_share_resources_third_parties">
            <?php foreach($options as $key => $name) {?>
            <option value=<?php echo $key?> <?php if ( $share_resources_third_parties == $key ) echo "selected" ?>><?php echo $name ?>
                <?php } ?>
        </select>
        <?php
    }

    /**
     * Insert option into settings form
     * When multisite is enabled, these options are not accessible here
     *
     * @since  2.0.0
     *
     * @access public
     *
     */

    public function get_option_enable_php_headers() {
        $rsssl_security_headers_method = $this->get_networkwide_option('rsssl_security_headers_method');
	    $options = [
		    'advancedheaders' => 'advanced-headers.php',
		    'htaccess' => '.htaccess',
	    ];
	    //keep this one for backward compatibility for users that have it enabled.
        if ( $rsssl_security_headers_method==='nginxconf' ) {
          $options['nginxconf'] = 'nginx.conf';
        }
	    ?>
        <select name="rsssl_security_headers_method" id="rsssl-security-headers-method">
            <?php foreach($options as $key => $name) { ?>
            <option  value=<?php echo $key?> <?php if ($rsssl_security_headers_method == $key) echo "selected" ?>><?php echo $name ?>
                <?php } ?>
        </select>
	    <?php
    }

	/**
	 * Insert option into settings form
	 * When multisite is enabled, these options are not accessible here
	 *
	 * @since  2.0.0
	 *
	 * @access public
	 *
	 */
	public function get_option_x_xss_protection() {
		$enabled = $this->get_networkwide_option('rsssl_x_xss_protection');
//        if ( $this->security_header_is_set_by_thirdparty('rsssl_x_xss_protection') ){
//	        $confirm = "3d ($enabled)";
//	        $enabled = true;
//        } else {
//	        $confirm = "NOT 3d ($enabled)";
//        }
		?>
        <label class="rsssl-switch">
            <input name="rsssl_x_xss_protection" size="40" value="1"
                   type="checkbox" <?php checked(1, $enabled, true) ?> />
            <span class="rsssl-slider rsssl-round"></span>
        </label>
		<?php
	}
	/**
	 * Insert option into settings form
	 * When multisite is enabled, these options are not accessible here
	 *
	 * @since  2.0.0
	 *
	 * @access public
	 *
	 */
	public function get_option_x_content_type_options() {
		$enabled = $this->get_networkwide_option('rsssl_x_content_type_options');
//        if ( $this->security_header_is_set_by_thirdparty('rsssl_x_content_type_options') ){
//	        $confirm = "3d ($enabled)";
//	        $enabled = true;
//        } else {
//	        $confirm = "NOT 3d ($enabled)";
//        }
		?>
        <label class="rsssl-switch">
            <input name="rsssl_x_content_type_options" size="40" value="1"
                   type="checkbox" <?php checked(1, $enabled, true) ?> />
            <span class="rsssl-slider rsssl-round"></span>
        </label>
		<?php
	}
	/**
	 * Insert option into settings form
	 * When multisite is enabled, these options are not accessible here
	 *
	 * @since  2.0.0
	 *
	 * @access public
	 *
	 */
	public function get_option_no_referrer_when_downgrade() {
		$enabled = $this->get_networkwide_option('rsssl_no_referrer_when_downgrade');
//        if ( $this->security_header_is_set_by_thirdparty('rsssl_no_referrer_when_downgrade') ){
//	        $confirm = "3d ($enabled)";
//	        $enabled = true;
//        } else {
//	        $confirm = "NOT 3d ($enabled)";
//        }
		?>
        <label class="rsssl-switch">
            <input name="rsssl_no_referrer_when_downgrade" size="40" value="1"
                   type="checkbox" <?php checked(1, $enabled, true) ?> />
            <span class="rsssl-slider rsssl-round"></span>
        </label>
		<?php
	}

	/**
	 * Insert option into settings form
	 * When multisite is enabled, these options are not accessible here
	 *
	 * @since  2.0.0
	 *
	 * @access public
	 *
	 */
	public function get_option_x_frame_options() {
		$enabled = $this->get_networkwide_option('rsssl_x_frame_options');
//        if ( $this->security_header_is_set_by_thirdparty('rsssl_x_frame_options') ){
//	        $confirm = "3d ($enabled)";
//	        $enabled = true;
//        } else {
//	        $confirm = "NOT 3d ($enabled)";
//        }
		?>
        <label class="rsssl-switch">
            <input name="rsssl_x_frame_options" size="40" value="1"
                   type="checkbox" <?php checked(1, $enabled, true) ?> />
            <span class="rsssl-slider rsssl-round"></span>
        </label>
		<?php
	}

	/**
	 * Insert option into settings form
	 * When multisite is enabled, these options take their value from the multisite settings if these are enabled
	 *
	 * @since  2.0.0
	 *
	 * @access public
	 *
	 */
	public function get_option_cert_expiration_warning() {
		$comment = "";
		$disabled = "";
		$cert_expiration_warning = get_option('rsssl_cert_expiration_warning');
		if ( is_multisite() && rsssl_multisite::this()->cert_expiration_warning) {
			$disabled = "disabled";
			$cert_expiration_warning = TRUE;
			$comment = __("This option is enabled on the network menu.", "really-simple-ssl-pro");
		}

		?>
        <label class="rsssl-switch">
            <input id="rlrsssl_options" name="rsssl_cert_expiration_warning" size="40" value="1"
                   type="checkbox" <?php echo $disabled?> <?php checked(1, $cert_expiration_warning, true) ?> />
            <span class="rsssl-slider rsssl-round"></span>
        </label>
		<?php
		RSSSL()->rsssl_help->get_comment($comment);
	}

	/**
	 * Insert option into settings form
	 * When multisite is enabled, these options are not accessible here
	 *
	 * @since  2.0.0
	 *
	 * @access public
	 *
	 */

	public function get_option_admin_mixed_content_fixer() {
		$admin_mixed_content_fixer = get_option('rsssl_admin_mixed_content_fixer');
		$disabled = "";
		$comment = "";

		if ( is_multisite() && RSSSL()->rsssl_multisite->mixed_content_admin ) {
			$disabled = "disabled";
			$admin_mixed_content_fixer = true;
		}

		?>
        <label class="rsssl-switch">
            <input id="rlrsssl_options" name="rsssl_admin_mixed_content_fixer" size="40" value="1" <?php echo $disabled?>
                   type="checkbox" <?php checked(1, $admin_mixed_content_fixer, true) ?> />
            <span class="rsssl-slider rsssl-round"></span>
        </label>
		<?php
		RSSSL()->rsssl_help->get_comment($comment);
	}

	/**
	 * Insert option into settings form
	 * When multisite is enabled, these options are not accessible here
	 *
	 * @since  2.0.0
	 *
	 * @access public
	 *
	 */

	public function get_option_hsts_preload() {
		$enabled = $this->get_networkwide_option('rsssl_hsts_preload');

//        if ( $this->security_header_is_set_by_thirdparty('rsssl_hsts_preload') ) {
//	        $confirm = "3d ($enabled)";
//	        $enabled = true;
//        } else {
//	        $confirm = "NOT 3d ($enabled)";
//        }
        ?>
        <label class="rsssl-switch" id="rsssl-maybe-highlight-hsts_preload">
            <input id="hsts_preload" class="hsts_preload" name="rsssl_hsts_preload" size="40" value="1"
                   type="checkbox" <?php checked(1, $enabled, true) ?> />
            <span class="rsssl-slider rsssl-round"></span>
        </label>
		<?php
		$link_start ='<a target="_blank" href="https://hstspreload.appspot.com/?domain='.$this->non_www_domain().'">';
		$link_close = "</a>";
		$comment = sprintf(__("After enabling this option, you have to %ssubmit%s your site. Please read the instructions on this page carefully before doing so.", "really-simple-ssl-pro"), $link_start, $link_close );
		RSSSL()->rsssl_help->get_comment($comment, 'hsts_preload_expl');
	}

	public function get_option_network_admin_only() {
		/**
		 * Placeholder function to be able to show a notice about network admin settings
		 */
	}

	/**
	 * Insert option into settings form
	 * When multisite is enabled, these options are not accessible here
	 * If .htaccess is not editable, the setting is disabled
	 *
	 * @since  2.0.0
	 *
	 * @access public
	 *
	 */
	public function get_option_turn_on_permissions_policy() {
        $enabled = $this->get_networkwide_option('rsssl_turn_on_permissions_policy');
//		if ( $this->security_header_is_set_by_thirdparty('rsssl_turn_on_permissions_policy') ){
//			$confirm = "3d ($enabled)";
//			$enabled = true;
//		} else {
//			$confirm = "NOT 3d ($enabled)";
//		}
		?>
        <label class="rsssl-switch">
            <input name="rsssl_turn_on_permissions_policy" size="40" value="1"
                   type="checkbox" <?php checked(1, $enabled, true) ?> />
            <span class="rsssl-slider rsssl-round"></span>
        </label>
		<?php
	}

	/**
	 * @return void
     * @since 5.4
     * Remove/add permissions' policy features
	 */
    public function update_permissions_policy_features() {

	    $permissions_policy = $this->get_networkwide_option('rsssl_permissions_policy');

	    $obsolete_features = array(
            'gyroscope',
            'magnetometer',
            'picture-in-picture',
            'sync-xhr',
            'usb',
            'interest-cohort',
        );

	    foreach ( $obsolete_features as $obsolete_feature ) {
            if ( isset ( $permissions_policy[$obsolete_feature] ) ) {
                unset ( $permissions_policy[$obsolete_feature]);
            }
        }

	    $new_features = array(
	        'display-capture' => '*',
        );

        foreach ( $new_features as $new_feature => $value ) {
            if ( ! isset( $permissions_policy[$new_feature] ) ) {
	            $permissions_policy[ $new_feature ] = $value;
            }
        }

        $this->update_networkwide_option('rsssl_permissions_policy', $permissions_policy );
    }
	/**
	 * Insert option into settings form
	 *
	 * @since  2.0.0
	 *
	 * @access public
	 *
	 */

	public function get_option_permissions_policy() {
		$permissions_policy_values = $this->get_networkwide_option('rsssl_permissions_policy');
		if ( !$permissions_policy_values ) {
			$this->set_permissions_policy_defaults();
			$permissions_policy_values = $this->get_networkwide_option('rsssl_permissions_policy');
		}
		$options = array(
			'all' => '*',
			'self' => 'self',
			'none' => 'none',
		);
        $permissions_policy_values = wp_parse_args($permissions_policy_values, $this->permissions_policy_options );
		?>
		<?php foreach( $permissions_policy_values as $option_key => $option_value ) {
			?>
            <tr class="permissions-policy-setting">
                <td class="permissions-policy-name"><?php echo $option_key ?></td>
				<?php foreach ( $options as $key => $value ) {
					?>
                    <td>
                        <input type="radio" name="permissions-policy-<?php echo $option_key ?>" value="<?php echo $value ?>" <?php if ($value == $option_value) echo "checked='checked'" ?>>
                        <label for="<?php echo $option_key ?>">
                    </td>
					<?php
				}
				?>
            </tr>
		<?php } ?>
		<?php
	}

	/**
	 * Save the Permissions Policy values
	 */

	public function save_permissions_policy() {
		if ( ! current_user_can('manage_options' ) ) {
            return;
        }

		if (isset($_POST['security_headers_update']) && wp_verify_nonce($_POST['security_headers_update'], 'submit_security_headers')) {
			$permissions_policy_values = $this->get_networkwide_option( 'rsssl_permissions_policy' );
			$permissions_policy_values = wp_parse_args($permissions_policy_values, $this->permissions_policy_options );

			if ( empty( $permissions_policy_values ) ) {
				return;
			}

			$safe_keys = array(
				'*',
				'self',
				'none',
			);

			foreach ( $permissions_policy_values as $option_key => $option_value ) {
				if ( isset( $_POST["permissions-policy-$option_key"] ) ) {
					if ( in_array( $_POST["permissions-policy-$option_key"], $safe_keys ) ) {
						$permissions_policy_values[ $option_key ] = $_POST["permissions-policy-$option_key"];
					}
				}
			}

			$this->update_networkwide_option( 'rsssl_permissions_policy', $permissions_policy_values );
		}
	}

	/**
     * Set default permissions policy values
     *
	 * @return void
	 */
    public function set_permissions_policy_defaults() {
        $values = array();
        foreach ( $this->permissions_policy_options as $option => $value ) {
	        $values[ $option ] = $value;
        }

	    $this->update_networkwide_option( 'rsssl_permissions_policy', $values );
    }

    /**
     * Get permissions policy rules
     * @param bool $html_output
     * @param string|bool $type
     * @return string
     */

    public function generate_permissions_policy_header( $html_output = false, $type = false) {
	    $permissions_policy_values = $this->get_networkwide_option('rsssl_permissions_policy');
	    $rules = '';
	    foreach ( $permissions_policy_values as $policy => $value ) {
		    switch ($value) {
			    case '*':
                    $rules .= $policy ."=(*)" . ", ";
				    break;
			    case 'none':
				    $rules .= $policy ."=()" .", ";
				    break;
			    case 'self':
				    $rules .= $policy ."=(self)" .", ";
				    break;
		    }
	    }
	    // Remove last space and , from string
	    $rules = substr_replace($rules ,"",-2);
	    $rule = $this->wrap_header('Permissions-Policy', $rules, $type, $html_output);
	    $php_rule = $this->wrap_header('Permissions-Policy', $rules, 'php');

        update_option('rsssl_pro_permissions_policy_headers_for_php', $php_rule);
        return $rule;
    }

	/**n
	 * @param string $header
	 * @param string $rules
	 * @param string $type
	 * @param bool $html_output
	 *
	 * @return string
	 */
    public function wrap_header( $header, $rules, $type = 'apache', $html_output = false ){
	    $append = '';
	    if ($html_output) {
		    $break = "<br>";
	    } else {
		    $break = "\n";
	    }

	    if ( $header === 'Strict-Transport-Security' ){
		    if ( $type === 'nginx' ) {
			    $append = ' always';
		    } else if ($type === 'apache' ) {
			    $append = ' env=HTTPS';
		    }
	    }

	    if ( $type==='nginx' ) {
		    //on nginx
		    //with colon
		    //with quotes
		    //preceded by add_header
		    $rule = 'add_header '.$header.': "'.$rules.'"'.$append.$break;
	    } else {
	        //on apache/htaccess
            //no colon
            //with quotes
            //preceded by header always set
		    $rule = 'Header always set '.$header.' "'.$rules.'" '.$append.$break;
	    }

	    return $rule;
    }

	/**
	 * check if pretty permalinks are enabled
	 *
	 * @return false
	 */
	public function pretty_permalinks_enabled_required(){
        if (rsssl_get_networkwide_option('rsssl_content_security_policy') !== 'report-only') {
            return false;
        }

        if ( get_option('permalink_structure') ) {
            return false;
        }

        return true;
	}


	public function non_www_domain(){
		$domain = get_home_url();
		return str_replace(array("https://", "http://", "https://www.", "http://www.", "www."), "", $domain);
	}

	/**
	 * Add settings link on plugins overview page
	 * @param array $links
	 *
	 * @return array
	 */

	public function plugin_settings_link($links) {

		if ( is_network_admin() ) {
			$link = add_query_arg( array('page' => 'really-simple-ssl' ), network_admin_url('settings.php') );
		} else {
			$link = add_query_arg( array('page' => 'rlrsssl_really_simple_ssl' ), admin_url('options-general.php') );
		}
		$settings_link = '<a href="'.$link.'">'.__("Settings","really-simple-ssl-pro").'</a>';

		array_unshift($links, $settings_link);
		return $links;
	}

	/**
	 * Generate security headers, and insert in .htaccess file
     * This function only runs on settings page, or during upgrade, forced.
     * During upgrade, it skips if headers check is active, so will run again later
     * When on settings page, the function will run again of itself when headers check is complete
     *
	 * @param bool $force
	 */

	public function insert_security_headers( $force = false ){
	    if ( !$force && !$this->is_settings_page() ) {
            return;
	    }

		if ( defined('rsssl_pp_version') ) {
            return;
		}

		if ( function_exists("rocket_clean_domain") ) {
		    return;
        }

		//we make sure the headers are generated, for php and nginx purposes.
		$rules = $this->generate_security_header_rules(false, 'apache' );
        if ( $this->get_networkwide_option('rsssl_security_headers_method') !== 'htaccess' ) {
            return;
        }

        if ( empty( $rules) ) {
            $this->remove_htaccess_rules('Really_Simple_SSL_SECURITY_HEADERS', $force );
        } else {
            $this->write_to_htaccess($rules, 'Really_Simple_SSL_SECURITY_HEADERS', $force);
        }
    }



	/**
     * Add rules to WP Rocket htaccess
	 * @param string $htaccess_rules
	 *
	 * @return string
	 */

    public function add_rules_before_wp_rocket($htaccess_rules) {
	    if ( wp_doing_ajax()
             || !$this->is_settings_page()
	         || !RSSSL()->really_simple_ssl->ssl_enabled
	         || !current_user_can("activate_plugins")
	         || RSSSL()->really_simple_ssl->do_not_edit_htaccess
	    ) return $htaccess_rules;

        $output = "# BEGIN Really_Simple_SSL_SECURITY_HEADERS\n";
        $output .= "# Really Simple SSL / WP Rocket compatibility mode\n";
        $output .= "<IfModule mod_headers.c>"."\n";
        $output .= $this->generate_security_header_rules();
        $output .= "</IfModule>"."\n";
        $output .= "# END Really_Simple_SSL_SECURITY_HEADERS";
        $htaccess_rules .= $output;

	    return $htaccess_rules;
    }

	/**
	 * Regenerate the wp rocket .htaccess rules
	 */
	public function maybe_flush_wprocket_htaccess(){
		if ( wp_doing_ajax()
		     || !$this->is_settings_page()
		     || !RSSSL()->really_simple_ssl->ssl_enabled
		     || !current_user_can("activate_plugins")
		     || RSSSL()->really_simple_ssl->do_not_edit_htaccess
		) return;

		if ( function_exists('flush_rocket_htaccess') ) {
			flush_rocket_htaccess();
		}

		if ( function_exists('rocket_generate_config_file') ) {
			rocket_generate_config_file();
		}

	}

    /**
     * Get the security headers rules. Use $html_output to get HTML output, set $type to nginx to show NGINX rules
     *
     * @param bool $html_output
     * @param string $type
     * @return string
     *
     */

	public function generate_security_header_rules( $html_output = false, $type = 'apache' ) {

		//Get values for each security header
		$hsts = $this->get_networkwide_option('rsssl_hsts');
		$x_xss_protection = $this->get_networkwide_option('rsssl_x_xss_protection');
		$upgrade_insecure_requests = $this->get_networkwide_option('rsssl_upgrade_insecure_requests');
		$x_content_type_options = $this->get_networkwide_option('rsssl_x_content_type_options');
		$no_referrer_when_downgrade = $this->get_networkwide_option('rsssl_no_referrer_when_downgrade');
		$x_frame_options = $this->get_networkwide_option('rsssl_x_frame_options');
		$permissions_policy = $this->get_networkwide_option('rsssl_turn_on_permissions_policy');
        $block_third_party_popups = $this->get_networkwide_option('rsssl_block_third_party_popups');
        $share_resources_third_parties = $this->get_networkwide_option('rsssl_share_resources_third_parties');

		$rule = '';
        if ( $hsts ) {
            //not adding  env=HTTPS causes errors on lots of servers.
            // Remove the HSTS header from the old block before adding it to the new block
            $hsts_preload = $this->get_networkwide_option("rsssl_hsts_preload");
            if ( $hsts_preload ){
	            $rule .= $this->wrap_header('Strict-Transport-Security', "max-age=63072000; includeSubDomains; preload", $type, $html_output);
            } else {
                $rule .= $this->wrap_header('Strict-Transport-Security', "max-age=31536000", $type, $html_output);
            }
        }

		// Do not add the upgrade-insecure-requests header here when CSP is enforced, CSP will include this option when it is enabled
		if ($this->get_networkwide_option('rsssl_content_security_policy') !== 'enforce' && $this->get_networkwide_option('rsssl_content_security_policy') !== 'report-only' ) {
			if ($upgrade_insecure_requests) {
				$rule .= $this->wrap_header('Content-Security-Policy', "upgrade-insecure-requests", $type, $html_output);
			}
		} else if ( $type !== 'apache' || $html_output) {
			$rule .= RSSSL_PRO()->rsssl_csp_backend->get_csp_rules( $type, $html_output );
		}

        if ( $x_xss_protection ) {
	        $rule .= $this->wrap_header('X-XSS-Protection', "0", $type, $html_output);
        }

        if ( $x_content_type_options ) {
	        $rule .= $this->wrap_header('X-Content-Type-Options', "nosniff", $type, $html_output);
        }

        if ( $no_referrer_when_downgrade ) {
	        $rule .= $this->wrap_header('Referrer-Policy', "strict-origin-when-cross-origin", $type, $html_output);
        }

		if ( $permissions_policy ) {
			$rule .= $this->generate_permissions_policy_header($html_output, $type);
		}

        if ( $x_frame_options ) {
	        $rule .= $this->wrap_header('X-Frame-Options', "SAMEORIGIN", $type, $html_output);
        }

        // Cors headers
        if ( $block_third_party_popups === 'no' ) {
	        $rule .= $this->wrap_header('Cross-Origin-Opener-Policy', 'same-origin-allow-popups', $type, $html_output);
        }

        if ( $block_third_party_popups === 'yes' ) {
	        $rule .= $this->wrap_header('Cross-Origin-Opener-Policy', 'same-site', $type, $html_output);
        }

        if ( $share_resources_third_parties === 'no' ) {
            $rule .= $this->wrap_header( 'Cross-Origin-Resource-Policy', 'same-origin', $type, $html_output);
        }

        if ( $share_resources_third_parties === 'yes' ) {
            $rule .= $this->wrap_header( 'Cross-Origin-Resource-Policy', 'cross-origin', $type, $html_output);
        }

        if ( $share_resources_third_parties === 'yes_own_domain' ) {
            $rule .= $this->wrap_header( 'Cross-Origin-Resource-Policy', 'same-site', $type, $html_output);
        }

        if ( $html_output ) {
            // Remove last line break to end </code> on same line
            $rule = preg_replace('~<br>(?!.*<br>)~', '', $rule);

            //add wrapper comments
	        $rule = '# BEGIN Really_Simple_SSL_SECURITY_HEADERS<br>' .
                    $rule . '<br>' .
                    '# END Really_Simple_SSL_SECURITY_HEADERS';
        }

        return $rule;
    }

	/**
     * Write rules to the .htaccess file
     *
	 * @param string $rules
	 * @param string $name
	 * @param bool $force
	 */

	public function write_to_htaccess( $rules, $name, $force=false ) {
	    //Do not update if this is not the RSSSL settings page
		if ( !$force && !$this->is_settings_page()) {
            return;
		}

		$htaccess_filename = RSSSL()->really_simple_ssl->htaccess_file();
		if (
                wp_doing_ajax()
		     || !RSSSL()->really_simple_ssl->ssl_enabled
		     || !current_user_can("manage_options")
             || !file_exists( $htaccess_filename )
             || !is_writable( $htaccess_filename )
             || RSSSL()->really_simple_ssl->do_not_edit_htaccess
             || $this->get_networkwide_option('rsssl_security_headers_method') !== 'htaccess'
             || RSSSL_PRO()->rsssl_headers->is_header_check_running()
        ) return;

		$htaccess = file_get_contents( $htaccess_filename );

		//wrap rules
        $output = "# BEGIN ".$name."\n";
        $output .= "<IfModule mod_headers.c>"."\n";
        $output .= $rules;
        $output .= "</IfModule>"."\n";
        $output .= "# END ".$name;
        $output = preg_replace("/\n+/","\n", $output);

		if ( strpos($htaccess, $name) !==false ){
			//replace existing set
			$htaccess = preg_replace("/#\s?BEGIN\s?$name.*?#\s?END\s?$name/s", $output, $htaccess);
		} else {
            //nothing yet, insert fresh set
            $wptag = "# BEGIN WordPress";
			$output = "\n".$output."\n";
            if (strpos($htaccess, $wptag) !== false) {
                $htaccess = str_replace($wptag, $output . $wptag, $htaccess);
            } else {
                $htaccess =  $output . $htaccess;
            }
		}
		$htaccess = str_replace("\n"."\n"."\n", "\n"."\n", $htaccess);
		file_put_contents( $htaccess_filename , $htaccess);
    }

    /**
     * Check if the .htaccess file contains security headers
     * @since 4.1
     */

    public function htaccess_contains_security_headers() {
        $htaccess_file = RSSSL()->really_simple_ssl->htaccess_file();
        if (file_exists($htaccess_file) ) {
            $htaccess = file_get_contents($htaccess_file);
            if (strpos($htaccess, 'Really_Simple_SSL_SECURITY_HEADERS') !== false) {
                return true;
            }
        }
        return false;
    }

    /**
     * Check if one of the security header options has been enabled
     * @since 4.1
     */

    public function security_header_enabled() {

        if ($this->get_networkwide_option('rsssl_upgrade_insecure_requests'  ) ||
            $this->get_networkwide_option('rsssl_x_xss_protection') ||
            $this->get_networkwide_option('rsssl_x_content_type_options') ||
            $this->get_networkwide_option('rsssl_no_referrer_when_downgrade') ||
            $this->get_networkwide_option('rsssl_x_frame_options') ||
            $this->get_networkwide_option('rsssl_turn_on_permissions_policy') ) {
            return true;
        } else {
            return false;
        }
    }

	public function show_scan_buttons_before_activation() {
		$result = RSSSL_PRO()->rsssl_scan->scan_completed_no_errors();
		$scan_link = add_query_arg(array('page'=>'rlrsssl_really_simple_ssl', 'tab' => 'premium', 'rsssl_start_scan' => 1 ), admin_url('options-general.php'));
		if ( $result !== "NEVER" ) {
			$btn_text = __( "Scan again", "really-simple-ssl-pro" );
		} else {
			$btn_text = __( "Scan for issues", "really-simple-ssl-pro" );
		}

		if ( $result == "COMPLETED" ) { ?>
            <div class="rsssl-scan-text-in-activate-notice"><p><?php _e( "You finished a scan without errors.", "really-simple-ssl-pro" ) ?></p></div>
		<?php } elseif ( $result == "NEVER" ) { ?>
            <div class="rsssl-scan-text-in-activate-notice">
                <p>
					<?php

					$activate_link = add_query_arg(
						array(
							'page' => 'rlrsssl_really_simple_ssl',
							'tab' => 'license'
						), admin_url('options-general.php'));
					$activate =  ' '.sprintf(__("%sActivate%s your license.", "really-simple-ssl-pro"), '<a href="'.$activate_link.'">', '</a>');

					if ( ! RSSSL_PRO()->rsssl_licensing->license_is_valid() ) {
						echo __( "No scan completed yet. Before migrating to SSL, you should do a scan. Activate your license to start a scan.", "really-simple-ssl-pro" ) . $activate;
					} else {
						$link_start = '<a href="' . $scan_link . '">';
						echo sprintf( __( "No scan completed yet. Before migrating to SSL, you should do a %sscan%s", "really-simple-ssl-pro" ), $link_start, "</a>" );
					} ?>
                </p>
            </div>
		<?php } else { ?>
            <div class="rsssl-scan-text-in-activate-notice">
                <p><?php _e( "Previous scan completed with issues", "really-simple-ssl-pro" ); ?></p>
            </div>
		<?php } ?>

        <div class="rsssl-scan-button" style="margin-right: 10px">
			<?php
			if ( RSSSL_PRO()->rsssl_licensing->license_is_valid() ) {
				echo '<a href="' . $scan_link . '" class="button button-secondary">'.$btn_text.'</a>';
			} else {
				echo '<button class="button button-secondary" disabled>'.$btn_text.'</button>';
			}
			?>
        </div>
		<?php
	}

	/**
     * Remove contents from the .htaccess file from a certain type
	 * @param string $name
     * @param bool $force
	 */

	public function remove_htaccess_rules($name, $force=false ) {
		//Do not update if this is not the RSSSL settings page
		if ( !$force && !$this->is_settings_page() ) {
			return;
		}

		$htaccess_filename = RSSSL()->really_simple_ssl->htaccess_file();
		if (
			 wp_doing_ajax()
		     || (!current_user_can("activate_plugins") && !defined('RSSSL_DOING_CSP'))
		     || !file_exists( $htaccess_filename )
		     || !is_writable( $htaccess_filename )
		     || RSSSL()->really_simple_ssl->do_not_edit_htaccess
		) {
		    return;
        }

		if ( RSSSL_PRO()->rsssl_headers->is_header_check_running() ){
            return;
		}

        $htaccess = file_get_contents( $htaccess_filename );
        $htaccess = preg_replace("/#\s?BEGIN\s?$name.*?#\s?END\s?$name/s", "", $htaccess);
		$htaccess = str_replace("\n"."\n"."\n", "\n"."\n", $htaccess);
		file_put_contents( $htaccess_filename, $htaccess);
	}

    /**
     * Remove old Feature Policy rules from .htaccess
     * @since 4.1
     */

	public function replace_feature_policy_rules() {

        if ( ! file_exists( RSSSL()->really_simple_ssl->htaccess_file() ) ) {
            return;
        }

        if ( RSSSL()->really_simple_ssl->do_not_edit_htaccess ) {
            return;
        }

        $htaccess = file_get_contents( RSSSL()->really_simple_ssl->htaccess_file() );
        if ( ! is_writable( RSSSL()->really_simple_ssl->htaccess_file() ) ) {
            return;
        }

        $pattern = '/Header always set Feature-Policy(.*?)(.*);/m';
        $replacement = '';
        preg_replace($pattern, $replacement, $htaccess);
        file_put_contents( RSSSL()->really_simple_ssl->htaccess_file(), $htaccess );
    
        $this->insert_security_headers(true);
    }

	/**
	 * Show notice CSP next step
	 */

	public function show_notice_csp_enabled_next_steps()
	{
		if ($this->get_networkwide_option('rsssl_content_security_policy') === 'report-only' ) {

			// If notice has been permanently dismissed, or notice has been temporarily dismissed for a week which hasn't passed, return
			if ($this->get_networkwide_option("rsssl_pro_csp_notice_next_steps_notice_dismissed") ) {
				return;
			}

			add_action('admin_print_footer_scripts', array($this, 'insert_csp_next_steps_dismiss'));

			$link_open = '<a target="_blank" href="https://really-simple-ssl.com/knowledge-base/how-to-use-the-content-security-policy-generator/">';
			$link_close = '</a>';
			$premium_tab = esc_url( admin_url("options-general.php?page=rlrsssl_really_simple_ssl&tab=premium") );

			ob_start();

			?>
            <p><?php _e("Follow these steps to complete the setup:", "really-simple-ssl-pro"); ?></p>
            <ul class="message-ul">
                <li class="rsssl-activation-notice-li"><div class="rsssl-bullet"></div><?php _e("Reporting is active and will gather Content Security Policy directives available under ‘Content Security Policy configuration’.", "really-simple-ssl-pro"); ?></li>
                <li class="rsssl-activation-notice-li"><div class="rsssl-bullet"></div><?php _e("Reporting will be paused after 20 requests to limit server load. You can always manually set reporting to paused, if needed.", "really-simple-ssl-pro"); ?></li>
                <li class="rsssl-activation-notice-li"><div class="rsssl-bullet"></div><?php _e("To restart reporting, set Content Security Policy to ‘reporting active’.", "really-simple-ssl-pro"); ?></li>
                <li class="rsssl-activation-notice-li"><div class="rsssl-bullet"></div><?php _e("When there a no new directives reported, you can continue with the configuration and set the Content Security Policy to ‘Enforce’.", "really-simple-ssl-pro"); ?></li>
                <li class="rsssl-activation-notice-li"><div class="rsssl-bullet"></div><?php printf(__("For more information about this security header, please follow this %slink%s.", "really-simple-ssl-pro"), $link_open, $link_close); ?></li>
            </ul>
			<?php
			$content = ob_get_clean();
			$class = "updated is-dismissible";
			$title = __("Content Security Policy reporting enabled", "really-simple-ssl-pro");
			echo RSSSL()->really_simple_ssl->notice_html( $class, $title, $content );
		}
	}

    /**
     * Insert CSP notice dismiss
     */

	public function insert_csp_next_steps_dismiss() {
		if (!$this->get_networkwide_option("rsssl_pro_csp_notice_next_steps_notice_dismissed") ) {
			$ajax_nonce = wp_create_nonce( "really-simple-ssl-dismiss" );
			?>
            <script type='text/javascript'>
                jQuery(document).ready(function($) {
                    $(".notice.updated.is-dismissible").on("click", ".notice-dismiss", function(event){
                        rsssl_dismiss_csp_notice('dismiss');
                    });
                    function rsssl_dismiss_csp_notice(type){
                        var data = {
                            'action': 'dismiss_csp_next_steps_notice',
                            'type' : type,
                            'security': '<?php echo $ajax_nonce; ?>'
                        };
                        $.post(ajaxurl, data, function (response) {});
                    }
                });
            </script>
			<?php
		}
	}

    /**
     * Callback for the CSP notice dismiss function
     */

	public function dismiss_csp_next_steps_notice_callback()
	{
		$type = isset($_POST['type']) ? $_POST['type'] : false;

		if ($type === 'dismiss') {
			$this->update_networkwide_option('rsssl_pro_csp_notice_next_steps_notice_dismissed', true);
		}

		wp_die(); // this is required to terminate immediately and return a proper response
	}

	/**
     * Check if this option is enabled by a third party
	 * @param $option
	 *            //@todo Header detection disabled, enable in future version
	 * @return bool
	 */
//    public function security_header_is_set_by_thirdparty( $option ){
//        $headers_set_by_thirdparty = $this->get_active_security_headers_set_by_thirdparty();
//	    if ( in_array($option, $headers_set_by_thirdparty) ){
//		     return true;
//	    }
//        return false;
//    }

    /**
     * Get active security headers not set by RSSSL
     * @return array header_name -> option_name
     *                           //@todo Header detection disabled, enable in future version
     */
//    public function get_active_security_headers_set_by_thirdparty() {
//        $found_not_enabled_by_rsssl = [];
//        // Get used headers
//        $used_headers = RSSSL_PRO()->rsssl_headers->get_detected_security_headers( 'used' );
//        error_log("USED headers ");
//        error_log(print_r($used_headers, true));
//	    $security_header_names = array_column($this->security_headers, 'name');
//	    error_log("security_header_names ");
//	    error_log(print_r($security_header_names, true));
//	    //if headers are set with php or advanced headers, we have detected them without RSSSL headers.
//        //this means that ALL headers in the used array are always third party.
//	    if ( rsssl_get_networkwide_option('rsssl_security_headers_method') === 'php' || rsssl_get_networkwide_option('rsssl_security_headers_method') === 'advancedheaders' ) {
//		    foreach ( $used_headers as $header_name => $v ) {
//			    $found_key = array_search( $header_name, $security_header_names);
//                error_log("found key for $header_name $found_key");
//			    if ( $found_key!==FALSE ) {
//                    error_log($this->security_headers[ $found_key ]['option_name']);
//                    $found_not_enabled_by_rsssl[ $header_name ] = $this->security_headers[ $found_key ]['option_name'];
//			    }
//		    }
//        } else {
//		    foreach ( $used_headers as $header_name => $v ) {
//			    $found_key = array_search( $header_name, $security_header_names);
//			    if ( $found_key!==FALSE ) {
//				    $option = $this->security_headers[ $found_key ]['option_name'];
//				    if ( !$this->get_networkwide_option( $option ) || $used_headers[$header_name]['count']>1 ) {
//					    $found_not_enabled_by_rsssl[ $header_name ] = $option;
//				    }
//			    }
//		    }
//        }
//
//	    return $found_not_enabled_by_rsssl;
//    }

    /**
     * Add notice(s) for headers with non-recommended values
     *
     * @return array
     * @todo Header detection disabled, enable in future version
     */

//    public function add_non_recommended_header_notices( $notices ) {
//        //currently disabled
//        return $notices;
//
//	    $non_recommended_values = array();
//	    if ( RSSSL_PRO()->rsssl_headers->is_header_check_running() ) {
//		    return $notices;
//	    }
//
//	    $used_headers = RSSSL_PRO()->rsssl_headers->get_detected_security_headers( 'used' );
//        $security_header_names = array_column($this->security_headers, 'name');
//	    foreach ( $used_headers as $header_name => $header_value ) {
//            error_log("used headers");
//            error_log($header_name);
//		    // Skip CORS && CSP
//		    if ( $header_name === 'Cross-Origin-Opener-Policy' || $header_name === 'Cross-Origin-Resource-Policy') continue;
//
//            $found_key = array_search($header_name, $security_header_names);
//            if (isset( $header_value['value'] )) {
//                error_log("header value isset");
//            } else {
//                error_log("header value not set");
//            }
//		    if ( $this->security_headers[ $found_key ]['option_value']!==false ) {
//			    error_log("has option value ");
//                error_log(print_r($this->security_headers[ $found_key ]['option_value'], true));
//		    } else {
//                error_log("not has option value ");
//		    }
//            if ($found_key){
//                error_log("found key ".$found_key);
//            } else {
//                error_log("not found key ");
//            }
//
//		    if ( isset( $header_value['value'] ) && $this->security_headers[ $found_key ]['option_value']!==false && $found_key ) {
//			    $option_value = $this->security_headers[ $found_key ]['option_value'];
//                if ( !is_array($option_value) ) $option_value = [$this->security_headers[ $found_key ]['option_value']];
//                error_log("Check recommended value for ".$header_name);
//                error_log("actual value");
//                error_log(print_r($header_value['value'], true));
//
//			    error_log("recommended value");
//			    error_log(print_r($option_value, true));
//
//			    if ( !in_array($header_value['value'], $option_value) )  {
//				    // Value is different from recommended value
//				    $non_recommended_values[ $header_name ] = $header_value['value'];
//			    }
//		    }
//	    }
//
//        foreach ( $non_recommended_values as $header => $value ) {
//            $notices[ 'wrong_value_'.$header ] = array(
//                'callback' => '_true_',
//                'score' => 5,
//                'output' => array(
//                    'true' => array(
//	                    'url' => "https://really-simple-ssl.com/how-to-find-where-security-headers-are-set/",
//	                    'msg' => sprintf(__("Header %s has been set to non-recommended value: %s.", "really-simple-ssl-pro"), $header, $value ),
//                        'icon' => 'open',
//                        'dismissible' => true
//                    ),
//                ),
//            );
//        }
//        if (rsssl_get_networkwide_option('rsssl_security_headers_method') === 'php' || rsssl_get_networkwide_option('rsssl_security_headers_method') === 'advancedheaders'){
//	        $duplicate_headers=[];
//            //for php and advanced headers, the used headers ONLY contains headers that are not set by RSSSL.
//            //if any of these headers is also enabled in RSSSL, this is a duplicate header.
//	        foreach ( $used_headers as $header_name => $header_value ) {
//		        // Skip CORS && CSP
//		        if ( $header_name === 'Cross-Origin-Opener-Policy' || $header_name === 'Cross-Origin-Resource-Policy') continue;
//                error_log("Set by third party. Check if $header_name is set by RSSSL as well. ");
//		        $found_key = array_search($header_name, $security_header_names);
//		        if ( $found_key !== FALSE ) {
//			        $option_name = $this->security_headers[ $found_key ]['option_name'];
//                    $option_value = $this->get_networkwide_option($option_name);
//			        if ( $this->option_is_enabled($option_value) ) {
//                        error_log("header $option_name is set by RSSSL");
//				        // Header is set both by third party and Really Simple SSL
//				        $duplicate_headers[ $header_name ] = $header_value['value'];
//			        } else {
//				        error_log("header $option_name is NOT set by RSSSL");
//			        }
//		        }
//	        }
//	        foreach ( $duplicate_headers as $header => $value ) {
//		        $notices[ 'duplicate_header_'.$header ] = array(
//			        'callback' => '_true_',
//			        'score' => 5,
//			        'output' => array(
//				        'true' => array(
//					        'url' => "https://really-simple-ssl.com/how-to-find-where-security-headers-are-set/",
//					        'msg' => sprintf(__("Duplicate Security Header detected: %s. This can cause conflicts between the two set headers.", "really-simple-ssl-pro"), $header ),
//					        'icon' => 'open',
//					        'dismissible' => false,
//				        ),
//			        ),
//		        );
//	        }
//        }
//        return $notices;
//    }

	/**
     * Centralized method to check for several possible ways in which a value can be enabled.
	 * @param $value
	 *
	 * @return bool
     *
	 */
    public function option_is_enabled($value){
        if ( !$value ) {
            return false;
        }

	    if ( $value==='report-paused' || $value==='disabled'){
            return false;
	    }
        return true;
    }

	/**
	 * Check if site uses one of the most common caching tools.
	 *
	 * @return mixed
	 */

	public function site_uses_cache($str=false){

	    // W3 Total Cache
		if ( function_exists('w3tc_flush_all') ) {
		    if ($str) {
		        return 'W3 Total Cache';
		    } else {
			    return true;
		    }
		}

		// WP Fastest Cache
		if ( class_exists('WpFastestCache') ) {
			if ($str) {
				return 'Wp Fastest Cache';
			} else {
				return true;
			}
		}

		// WP Rocket
		if ( function_exists("rocket_clean_domain") ) {
			if ($str) {
				return 'WP Rocket';
			} else {
				return true;
			}
		}

		// WP Optimize
		if ( defined('WPO_PLUGIN_MAIN_PATH') ) {
			if ($str) {
				return 'WP Optimize';
			} else {
				return true;
			}
        }

		// WP Super Cache
        if ( defined('WPCACHEHOME') ) {
	        if ($str) {
		        return 'WP Super Cache';
	        } else {
		        return true;
	        }
        }

        // Hummingbird
        if ( defined('WPHB_VERSION') ) {
	        if ($str) {
		        return 'Hummingbird';
	        } else {
		        return true;
	        }
        }

        // Litespeed cache
        if ( defined('LSCWP_V') ) {
	        if ($str) {
		        return 'Litespeed cache';
	        } else {
		        return true;
	        }
        }

        // Autoptimize
        if ( defined('AUTOPTIMIZE_PLUGIN_VERSION') ) {
	        if ($str) {
		        return 'Autoptimize';
	        } else {
		        return true;
	        }
        }

        // Cache enabler
        if ( defined('CE_VERSION') ) {
	        if ($str) {
		        return 'Cache enabler';
	        } else {
		        return true;
	        }
        }

		return false;
	}

	public function maybe_clear_certificate_check_schedule( $old_value, $value, $option) {
		if ( $old_value==$value ) {
			return;
		}

		if ( !get_option('rsssl_cert_expiration_warning') ){
			wp_clear_scheduled_hook('rsssl_pro_daily_hook');
		}
	}

	/**
	 * Check to see if we are on the settings page, action hook independent
	 *
	 * @since  2.5
	 *
	 * @access public
	 *
	 */

	public function is_settings_page()
	{
		if (isset($_GET["page"]) && ($_GET["page"] == "rlrsssl_really_simple_ssl" || $_GET["page"] == "really-simple-ssl") ) {
			return true;
		}

		if (isset($_POST["option_page"]) && $_POST["option_page"] == "rlrsssl_pro_options" ) {
			return true;
		}

		return false;
	}

	/**
	 * Check if we can apply a setting network wide
	 */

	public function apply_networkwide_ssl_feature(){

		//if single site, always apply
		if (!is_multisite()) return true;

		//if multisite, only apply if we're on the network admin
		if ( is_multisite() && is_network_admin() ) return true;

		return false;
	}

}//class closure

if (!function_exists('rsssl_not_enabled_networkwide')) {
	function rsssl_not_enabled_networkwide() {
		if(is_multisite() && !RSSSL()->rsssl_multisite->plugin_network_wide_active() ){
			return true;
		}
		return false;
	}
}

if (!function_exists('rsssl_pro_hsts_enabled')) {
	function rsssl_pro_hsts_enabled() {
		return RSSSL_PRO()->rsssl_premium_options->get_networkwide_option( 'rsssl_hsts' );
	}
}

if (!function_exists('rsssl_security_headers_wpconfig_not_writable') ) {
    function rsssl_security_headers_wpconfig_not_writable()
    {
        if ( rsssl_get_networkwide_option('rsssl_security_headers_method') !== 'advancedheaders') {
            return false;
        }

        $wpconfig = RSSSL()->really_simple_ssl->find_wp_config_path();;
        if ( !$wpconfig ) {
            return true;
        }

        $wpconfig_contents = file_get_contents($wpconfig);
        if (strpos($wpconfig_contents, 'advanced-headers.php') !== false ) {
            return false;
        }

        if ( is_writable($wpconfig)
        ) {
            return false;
        }
        return true;
    }
}

if ( !function_exists('rsssl_pro_hsts_preload')) {
	function rsssl_pro_hsts_preload() {
		if ( ( RSSSL_PRO()->rsssl_premium_options->get_networkwide_option('rsssl_hsts')
             && RSSSL_PRO()->rsssl_premium_options->get_networkwide_option( 'rsssl_hsts_preload' )
            )
			//@todo Header detection disabled, enable in future version
             //|| ( RSSSL_PRO()->rsssl_premium_options->security_header_is_set_by_thirdparty('rsssl_hsts_preload') )
        ) {
			return true;
		}

		return false;
	}
}

if (!function_exists('rsssl_pro_check_for_new_csp_entries')) {
	function rsssl_pro_check_for_new_csp_entries() {

	    if (RSSSL_PRO()->rsssl_premium_options->get_networkwide_option('rsssl_content_security_policy') === 'report-paused') return 'report-paused';

		global $wpdb;

		$table_name = $wpdb->base_prefix . "rsssl_csp_log";
		//Check if there are any inpolicy values that are not true. If so, new rules can be added to the Content Security Policy. Show a warning in dashboard when new rules can be added, if all rules have been added show a checkmark
		$count = $wpdb->get_var( "SELECT count(*) FROM $table_name where inpolicy != 'true'" );

		if ( $count > 0 ) {
			return 'new-csp-rules';
		}

		return 'no-new-csp-rules';
	}
}

if (!function_exists('rsssl_pro_admin_mixed_content_fixer')) {
	function rsssl_pro_admin_mixed_content_fixer() {
		/*  Display the current settings for the admin mixed content. */
		$admin_mixed_content_fixer = get_option( "rsssl_admin_mixed_content_fixer" );

		if ( $admin_mixed_content_fixer ) {
			return 'admin-mixed-content-fixer-activated';
		}

		return 'admin-mixed-content-fixer-not-activated';
	}
}

if ( !function_exists('rsssl_pro_scan_notice') ) {
	function rsssl_pro_scan_notice() {
		if ( ! RSSSL()->really_simple_ssl->site_has_ssl ) {
			if ( RSSSL_PRO()->rsssl_scan->scan_completed_no_errors() == "COMPLETED" ) {
				return 'has-ssl-no-scan-errors';
			} elseif ( RSSSL_PRO()->rsssl_scan->scan_completed_no_errors() == "ERRORS" ) {
				return 'has-ssl-scan-has-errors';
			} elseif ( ! RSSSL_PRO()->rsssl_licensing->license_is_valid() ) {
                return 'no-ssl-no-scan-done-invalid-license';
            } else {
				return 'no-scan-done';
			}
		} else {
			if ( RSSSL_PRO()->rsssl_scan->scan_completed_no_errors() == "COMPLETED" ) {
				return 'no-ssl-no-scan-errors';
			} elseif ( RSSSL_PRO()->rsssl_scan->scan_completed_no_errors() == "ERRORS" ) {
				return 'no-ssl-scan-has-errors';
			} elseif ( ! RSSSL_PRO()->rsssl_licensing->license_is_valid() ) {
				return 'no-ssl-no-scan-done-invalid-license';
			} else {
				return 'no-scan-done';
			}
		}
	}
}

if (!function_exists('rsssl_pro_is_license_expired')) {
	function rsssl_pro_is_license_expired() {
		$status = RSSSL_PRO()->rsssl_licensing->get_license_status();
		if ( ! $status || $status === 'empty' || $status === 'site_inactive' || $status === 'deactivated' || $status === 'inactive' ) {
			return 'not-activated';
		} else if ( $status === 'revoked' || $status === 'missing' || $status ==='item_name_mismatch') {
		    return 'expired';
		} else if ($status === 'no_activations_left') {
		    return 'no_activations_left';
		} else {
		    return $status;
        }

	}
}

if (!function_exists('rsssl_redirect_to_homepage')) {
	function rsssl_redirect_to_homepage() {
		$redirect_to_homepage = RSSSL_PRO()->rsssl_premium_options->redirects_to_homepage();

		if ( $redirect_to_homepage == true ) {
			return 'redirect-to-homepage';
		}
	}
}

if ( !function_exists('rsssl_free_csp_incompatible') ) {
    function rsssl_free_csp_incompatible() {
        if ( version_compare( rsssl_version, '4.0.8', '<' ) ) {
            return true;
        }
        return false;
    }
}

if ( !function_exists('rsssl_csp_enabled') ) {
    function rsssl_csp_enabled() {
        if ( RSSSL_PRO()->rsssl_premium_options->get_networkwide_option('rsssl_content_security_policy') === 'report-only' ||
            RSSSL_PRO()->rsssl_premium_options->get_networkwide_option('rsssl_content_security_policy') === 'enforce' ) {
            return true;
        }
        return false;
    }
}

if ( !function_exists('rsssl_free_incompatible') ) {
	function rsssl_free_incompatible() {
		if ( version_compare( rsssl_version, '5.2.2', '<' ) ) {
			return true;
		}
		return false;
	}
}


if ( ! function_exists('rsssl_x_xss_protection_status') ) {
	/**
     *  When disabled by user, X-XSS-Protection is disabled and not set by third party, we show a notice that XSS should be enabled with setting 0
     *
	 * @return false|string
	 */

	function rsssl_x_xss_protection_status() {
        if ( get_option('rsssl_xss_protection_disabled_by_user') ) {
			return 'x-xss-disabled';
		}

		return false;
	}
}
//@todo Header detection disabled, enable in future version
//if ( !function_exists('rsssl_security_headers_disabled_by_user_set_by_thirdparty_condition')){
//    function rsssl_security_headers_disabled_by_user_set_by_thirdparty_condition(){
//        $h = rsssl_security_headers_disabled_by_user_set_by_thirdparty();
//        return !empty($h);
//    }
//}

/**
 * Check which headers are still enabled after disabling the option
 *             //@todo Header detection disabled, enable in future version
 */
//if ( ! function_exists('rsssl_security_headers_disabled_by_user_set_by_thirdparty') ) {
//	function rsssl_security_headers_disabled_by_user_set_by_thirdparty() {
//		// When disabled and header has been set by host/server
//		$disabled_by_user_but_set = get_option('rsssl_security_headers_set_by_host_disabled_by_user', []);
//		$headers_set_by_thirdparty = RSSSL_PRO()->rsssl_premium_options->get_active_security_headers_set_by_thirdparty();
//        $headers_still_enabled = array();
//        foreach ( RSSSL_PRO()->rsssl_premium_options->security_headers as $header ) {
//            if ( array_key_exists( $header['name'], $headers_set_by_thirdparty )
//                 && ( $disabled_by_user_but_set && in_array( $header['option_name'], $disabled_by_user_but_set ) )
//                 && ! in_array( $header['name'], $headers_still_enabled ) ) {
//	                $headers_still_enabled[] = $header['name'];
//            }
//        }
//
//		return $headers_still_enabled;
//	}
//}

if ( ! function_exists('rsssl_upgraded_to_current_version') ) {
	function rsssl_upgraded_to_current_version() {
		$current_version = get_option( 'rsssl_first_version' );
		//if there's no first version yet, we assume it's not upgraded
		if ( !$current_version ) {
			return false;
		}
		//if the first version is below current, we just upgraded.
		if ( version_compare($current_version,rsssl_pro_version ,'<') ){
			return true;
		}
		return false;
	}
}

if ( ! function_exists('rsssl_wp_content_not_writable') ) {
    function rsssl_wp_content_not_writable() {
        if (RSSSL_PRO()->rsssl_premium_options->get_networkwide_option( 'rsssl_security_headers_method' ) !== 'advancedheaders' ) {
            return false;
	    }
        if ( file_exists(WP_CONTENT_DIR.'/advanced-headers.php') && is_writable(WP_CONTENT_DIR.'/advanced-headers.php')) {
            return false;
        }
	    if ( ! is_writable( WP_CONTENT_DIR ) ) {
            return true;
        }
        if ( !is_writable(WP_CONTENT_DIR.'/advanced-headers.php') ) {
            return true;
        }

        return false;
    }
}